"""Classes used for steps in routers."""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from pydantic import ConfigDict, RootModel, model_validator
from typing_extensions import Self

if TYPE_CHECKING:
    from .manhattan import ManhattanRouterSide


@dataclass
class Step(ABC):
    """Abstract base for a routing step."""

    @abstractmethod
    def execute(self, router: ManhattanRouterSide, include_bend: bool) -> None:
        """Executes the step."""
        ...

    @property
    @abstractmethod
    def include_bend(self) -> bool | None:
        """Whether the execute should leave space at the end of the step for a bend."""
        ...


@dataclass
class Left(Step):
    """Let the router go left.

    If a size is given, the router will continues that amount
    (including the added bend).
    If include_bend is true, the router will stop before the next bend needs to bed
    placed at size's point.
    """

    dist: int | None = None
    include_bend: bool | None = None

    def execute(self, router: ManhattanRouterSide, include_bend: bool) -> None:
        """Make the router turn left and go straight if necessary."""
        _ib = include_bend if self.include_bend is None else self.include_bend
        router.left()
        if self.dist:
            if _ib:
                if (
                    self.dist is not None
                    and self.dist < 2 * router.router.bend90_radius
                ):
                    bend_radius = router.router.bend90_radius
                    raise ValueError(
                        "If the next bend should be avoided the step needs to be bigger"
                        f" than {2 * bend_radius=}"
                    )
                router.straight_nobend(self.dist)
            else:
                if self.dist < router.router.bend90_radius:
                    bend_radius = router.router.bend90_radius
                    raise ValueError(f"The step needs to be bigger than {bend_radius=}")
                router.straight(self.dist)


@dataclass
class Right(Step):
    """Let the router go right.

    If a size is given, the router will continues that amount
    (including the added bend).
    If include_bend is true, the router will stop before the next bend needs to bed
    placed at size's point.
    """

    dist: int | None = None
    include_bend: bool | None = None

    def execute(self, router: ManhattanRouterSide, include_bend: bool) -> None:
        """Adds the bend and potential straight after."""
        _ib = include_bend if self.include_bend is None else self.include_bend
        router.right()
        if self.dist:
            if _ib:
                if self.dist < 2 * router.router.bend90_radius:
                    bend_radius = router.router.bend90_radius
                    raise ValueError(
                        "If the next bend should be avoided the step needs to be bigger"
                        f" than {2 * bend_radius=}"
                    )
                router.straight_nobend(self.dist)
            else:
                if self.dist < router.router.bend90_radius:
                    bend_radius = router.router.bend90_radius
                    raise ValueError(f"The step needs to be bigger than {bend_radius=}")
                router.straight(self.dist)


@dataclass
class Straight(Step):
    """Adds a straight section to the router."""

    dist: int | None = None
    include_bend: bool | None = None

    def execute(self, router: ManhattanRouterSide, include_bend: bool) -> None:
        """Adds a straight section to the router."""
        _ib = include_bend if self.include_bend is None else self.include_bend
        if self.dist:
            if _ib:
                router.straight_nobend(self.dist)
            else:
                router.straight(self.dist)


@dataclass
class X(Step):
    """Go to an absolute X coordinate."""

    x: int
    include_bend: bool | None = None

    def execute(self, router: ManhattanRouterSide, include_bend: bool) -> None:
        """Adds a straight section to the router."""
        _ib = include_bend if self.include_bend is None else self.include_bend
        if router.t.angle % 2:
            raise ValueError(
                "Cannot go to position {self.x=}, because the router is currently "
                "going in the y direction with position "
                f"{(router.t.disp.x, router.t.disp.y)}"
            )
        if self.x:
            if _ib:
                router.straight_nobend(self.x - router.t.disp.x)
            else:
                router.straight(self.x - router.t.disp.x)


@dataclass
class Y(Step):
    """Go to an absolute X coordinate."""

    y: int
    include_bend: bool | None = None

    def execute(self, router: ManhattanRouterSide, include_bend: bool) -> None:
        """Adds a straight section to the router."""
        _ib = include_bend if self.include_bend is None else self.include_bend
        if router.t.angle % 2 == 0:
            raise ValueError(
                "Cannot go to position {self.x=}, because the router is currently "
                "going in the y direction with position "
                f"{(router.t.disp.x, router.t.disp.y)}"
            )
        if self.y:
            if _ib:
                router.straight_nobend(self.y - router.t.disp.y)
            else:
                router.straight(self.y - router.t.disp.y)


@dataclass
class XY(Step):
    """Go to an absolute XY coordinate."""

    x: int
    y: int
    include_bend: bool | None = None

    def execute(self, router: ManhattanRouterSide, include_bend: bool) -> None:
        """Executes the step on a router."""
        _ib = include_bend if self.include_bend is None else self.include_bend
        dx = self.x - router.t.disp.x
        dy = self.y - router.t.disp.y
        a = router.t.angle
        match a:
            case 0 | 2 if self.y == router.t.disp.x:
                sign = -1 if a == 2 else 1
                if sign * dx < 0:
                    raise ValueError(
                        "XY step cannot go back. It is current pointing at 0"
                        " degrees.\n"
                        f"Current position: {router.t.disp!r}.\n"
                        f"Target Position ({self.x},{self.y})"
                    )
                    if _ib:
                        router.straight_nobend(abs(dx))
                    else:
                        router.straight(abs(dx))
                else:
                    if sign * dx < router.t.disp.x + router.router.bend90_radius:
                        raise ValueError("XY step cannot go back")
                    router.straight_nobend(abs(dx))
                    if self.y > router.t.disp.y:
                        if a == 0:
                            router.left()
                        else:
                            router.right()
                    dy = self.y - router.t.disp.y
                    if _ib:
                        if abs(dy) < 0:
                            raise ValueError(
                                "XY's y-step is too small. It is current pointing"
                                f" at {router.t.angle * 90}"
                                " degrees.\n"
                                f"Current position: {router.t.disp!r}.\n"
                                f"Target Position ({self.x},{self.y})"
                            )
                        router.straight(abs(dy))
                    else:
                        if abs(dy) < router.router.bend90_radius:
                            raise ValueError(
                                "XY's y-step is too small. It is current pointing"
                                f" at {router.t.angle * 90}"
                                " degrees.\n"
                                f"Current position: {router.t.disp!r}.\n"
                                f"Target Position ({self.x},{self.y})\n"
                                "Too small distance to place bend of "
                                f"{router.router.bend90_radius} size"
                            )
                        router.straight_nobend(abs(dy))

            case 1 | 3 if self.x == router.t.disp.x:
                sign = -1 if a == 3 else 1
                if sign * dy < 0:
                    raise ValueError(
                        "XY step cannot go back. It is current pointing at 0"
                        " degrees.\n"
                        f"Current position: {router.t.disp!r}.\n"
                        f"Target Position ({self.x},{self.y})"
                    )
                    if _ib:
                        router.straight_nobend(abs(dy))
                    else:
                        router.straight(abs(dy))
                else:
                    if sign * dy < router.t.disp.y + router.router.bend90_radius:
                        raise ValueError("XY step cannot go back")
                    router.straight_nobend(abs(dx))
                    if self.x > router.t.disp.x:
                        if a == 3:
                            router.left()
                        else:
                            router.right()
                    dx = self.x - router.t.disp.x
                    if _ib:
                        if abs(dy) < 0:
                            raise ValueError(
                                "XY's y-step is too small. It is current pointing"
                                f" at {router.t.angle * 90}"
                                " degrees.\n"
                                f"Current position: {router.t.disp!r}.\n"
                                f"Target Position ({self.x},{self.y})"
                            )
                        router.straight(abs(dy))
                    else:
                        if abs(dy) < router.router.bend90_radius:
                            raise ValueError(
                                "XY's y-step is too small. It is current pointing"
                                f" at {router.t.angle * 90}"
                                " degrees.\n"
                                f"Current position: {router.t.disp!r}.\n"
                                f"Target Position ({self.x},{self.y})\n"
                                "Too small distance to place bend of "
                                f"{router.router.bend90_radius} size"
                            )
                        router.straight_nobend(abs(dy))


class Steps(RootModel[list[Any]]):
    """Collection of steps. Runs the execution on them."""

    root: list[Any]

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @model_validator(mode="after")
    def _check_steps(self) -> Self:
        for step in self.root:
            if not isinstance(step, Step):
                raise ValueError(
                    "All Steps must implement an "
                    "'execute(self, router: ManhattanRouterSide, include_bend: bool)"
                )
        return self

    def execute(self, router: ManhattanRouterSide) -> None:
        """Run all the steps on the given router."""
        try:
            # print(router.t)
            i = 0
            if self.root:
                for i, step in enumerate(self.root[:-1]):
                    step.execute(router, True)
                i += i
                step = self.root[-1]
                step.execute(router, False)
            # print(router.t)

        except Exception as e:
            raise ValueError(f"Error in step {i}, {step=}. Error: {str(e)}") from e



"""Utilities to fix DRC violations.

:py:func:~`fix_spacing_tiled` uses :py:func:~`kdb.Region.space_check` to detect
minimum space violations and then applies a fix.
"""

from typing import overload

from .. import KCell, kdb
from ..conf import config, logger

__all__ = [
    "fix_spacing_tiled",
    "fix_spacing_sizing_tiled",
    "fix_spacing_minkowski_tiled",
]


@overload
def fix_spacing_tiled(
    c: KCell,
    min_space: int,
    layer: kdb.LayerInfo,
    metrics: kdb.Metrics = kdb.Metrics.Euclidian,
    ignore_angle: float = 80,
    size_space_check: int = 5,
    n_threads: int = 4,
    tile_size: tuple[float, float] | None = None,
    overlap: float = 3,
    smooth_factor: float = 0.05,
) -> kdb.Region: ...


@overload
def fix_spacing_tiled(
    c: KCell,
    min_space: int,
    layer: kdb.LayerInfo,
    metrics: kdb.Metrics = kdb.Metrics.Euclidian,
    ignore_angle: float = 80,
    size_space_check: int = 5,
    n_threads: int = 4,
    tile_size: tuple[float, float] | None = None,
    overlap: float = 3,
    *,
    smooth_absolute: int,
) -> kdb.Region: ...


def fix_spacing_tiled(
    c: KCell,
    min_space: int,
    layer: kdb.LayerInfo,
    metrics: kdb.Metrics = kdb.Metrics.Euclidian,
    ignore_angle: float = 80,
    size_space_check: int = 5,
    n_threads: int | None = None,
    tile_size: tuple[float, float] | None = None,
    overlap: float = 3,
    smooth_factor: float = 0.05,
    smooth_absolute: int | None = None,
    smooth_keep_hv: bool = True,
) -> kdb.Region:
    """Fix minimum space violations.

    Fix min space issues by running a drc check on the input region and merging
    it with the affcted polygons.

    Args:
        c: Input cell
        min_space: Minimum space rule [dbu]
        layer: Input layer index
        metrics: The metrics to use to determine the violation edges
        ignore_angle: ignore angles greater or equal to this angle
        size_space_check: Sizing in dbu of the offending edges towards the polygons
        n_threads: on how many threads to run the check simultaneously
        tile_size: tuple determining the size of each sub tile (in um), should be big
            compared to the violation size
        overlap: how many times bigger to make the tile border in relation to the
            violation size. Smaller than 1 can lead to errors
        smooth_factor: how big to smooth the resulting region in relation to the
            violation. 1 == the violation size. Set to 0 to disable
        smooth_absolute: If set will overwrite smooth with an an absolute value, not
            relative to the violation size. If set, this will disable smooth_factor.
            [dbu]
        smooth_keep_hv: Keep horizontal and vertical vertices when smoothing.

    Returns:
        fix: Region containing the fixes for the violations

    """
    if tile_size is None:
        min(25 * min_space, 250)
        tile_size = (30 * min_space * c.kcl.dbu, 30 * min_space * c.kcl.dbu)
    li = c.kcl.find_layer(layer)
    tp = kdb.TilingProcessor()
    tp.frame = c.kcl.to_um(c.bbox(li))  # type: ignore[misc]
    tp.dbu = c.kcl.dbu
    tp.tile_size(*tile_size)  # tile size in um
    tp.tile_border(min_space * overlap * tp.dbu, min_space * overlap * tp.dbu)
    tp.input("reg", c.kcl.layout, c.cell_index(), li)
    tp.threads = n_threads or config.n_threads

    fix_reg = RegionOperator()
    tp.output("fix_reg", fix_reg)

    if smooth_factor != 0 or smooth_absolute:
        keep = "true" if smooth_keep_hv else "false"
        smooth = (
            min(int(smooth_factor * min_space), 1)
            if not smooth_absolute
            else smooth_absolute
        )
        queue_str = (
            f"var sc = reg.space_check({min_space},"
            f" false, Metrics.{metrics.to_s()},"
            f" {ignore_angle}); "
            "var edges = sc.edges(); edges.merge(); "
            f"var r_int = (edges.extended(0, 0, 0, {size_space_check}, true)"
            " + sc.polygons()); r_int.merge();"
            " r_int.insert(reg.interacting(sc.polygons())); "
            f"r_int.merge(); r_int.smooth({smooth}, {keep}); "
            f"_output(fix_reg, r_int)"
        )
    else:
        queue_str = (
            f"var sc = reg.space_check({min_space},"
            f" false, Metrics.{metrics.to_s()},"
            f" {ignore_angle});"
            "var edges = sc.edges(); edges.merge();"
            f"var r_int = (edges.extended(0, 0, 0, {size_space_check}, true)"
            " + sc.polygons()); r_int.merge();"
            " r_int.insert(reg.interacting(sc.polygons()));"
            "r_int.merge(); _output(fix_reg, r_int)"
        )

    tp.queue(queue_str)

    c.kcl.start_changes()
    tp.execute("Min Space Fix")
    c.kcl.end_changes()

    return fix_reg.region


def fix_spacing_sizing_tiled(
    c: KCell,
    min_space: int,
    layer: kdb.LayerInfo,
    n_threads: int | None = None,
    tile_size: tuple[float, float] | None = None,
    overlap: int = 2,
) -> kdb.Region:
    """Fix min space issues by using a dilation & erosion.

    Args:
        c: Input cell
        min_space: Minimum space rule [dbu]
        layer: Input layer
        n_threads: on how many threads to run the check simultaneously
        tile_size: tuple determining the size of each sub tile (in um), should be big
            compared to the violation size
        overlap: how many times bigger to make the tile border in relation to the
            violation size. Smaller than 1 can lead to errors

    Returns:
        kdb.Region: Region containing the fixes for the violations

    """
    tp = kdb.TilingProcessor()
    if tile_size is None:
        size = min_space * 20 * c.kcl.dbu
        tile_size = (size, size)
    li = c.kcl.find_layer(layer)
    tp.frame = c.kcl.to_um(c.bbox(li))  # type: ignore[misc]
    tp.dbu = c.kcl.dbu
    tp.tile_size(*tile_size)  # tile size in um
    tp.tile_border(min_space * overlap * tp.dbu, min_space * overlap * tp.dbu)
    tp.input("reg", c.kcl.layout, c.cell_index(), li)
    tp.threads = n_threads or config.n_threads

    fix_reg = kdb.Region()

    tp.output("fix_reg", fix_reg)

    queue_str = (
        "var tile_reg=reg & (_tile & _frame);"
        + f"reg = tile_reg.sized({min_space}).sized({-min_space});"
        + "_output(fix_reg, reg)"
    )

    tp.queue(queue_str)

    c.kcl.start_changes()
    tp.execute("Min Space Fix")
    c.kcl.end_changes()

    return fix_reg


def fix_spacing_minkowski_tiled(
    c: KCell,
    min_space: int,
    ref: kdb.LayerInfo | kdb.Region,
    n_threads: int | None = None,
    tile_size: tuple[float, float] | None = None,
    overlap: int = 1,
    smooth: int | None = None,
) -> kdb.Region:
    """Fix min space issues by using a dilation & erosion with a box.

    Args:
        c: Input cell
        min_space: Minimum space rule [dbu]
        ref: Input layer index or region
        n_threads: on how many threads to run the check simultaneously
        tile_size: tuple determining the size of each sub tile (in um), should be big
            compared to the violation size
        overlap: how many times bigger to make the tile border in relation to the
            violation size. Smaller than 1 can lead to errors
        smooth: Apply smoothening (simplifying) at the end if > 0

    Returns:
        kdb.Region: Region containing the fixes for the violations
    """
    tp = kdb.TilingProcessor()
    tp.frame = c.dbbox()  # type: ignore[misc]
    tp.dbu = c.kcl.dbu
    tp.threads = n_threads or config.n_threads

    min_tile_size_rec = 10 * min_space * tp.dbu

    if tile_size is None:
        tile_size = (min_tile_size_rec * 2, min_tile_size_rec * 2)

    tp.tile_border(min_space * overlap * tp.dbu, min_space * overlap * tp.dbu)

    tp.tile_size(*tile_size)
    if isinstance(ref, kdb.LayerInfo):
        tp.input("main_layer", c.kcl.layout, c.cell_index(), c.kcl.find_layer(ref))
    else:
        tp.input("main_layer", ref)

    operator = RegionOperator()
    tp.output("target", operator)
    if smooth is None:
        queue_str = (
            f"var tile_reg = (_tile & _frame).sized({min_space});"
            f"var shape = Box.new({min_space},{min_space});"
            "var reg = main_layer.minkowski_sum(shape); reg.merge();"
            "reg = tile_reg - (tile_reg - reg).minkowski_sum(shape);"
            "_output(target, reg & _tile, true);"
        )
    else:
        queue_str = (
            f"var tile_reg = (_tile & _frame).sized({min_space});"
            f"var shape = Box.new({min_space},{min_space});"
            "var reg = main_layer.minkowski_sum(shape); reg.merge();"
            "reg = tile_reg - (tile_reg - reg).minkowski_sum(shape);"
            f"reg.smooth({smooth});"
            "_output(target, reg & _tile, true);"
        )

    tp.queue(queue_str)
    logger.debug("String queued for {}:  {}", c.name, queue_str)

    c.kcl.start_changes()
    logger.info("Starting minkowski on {}", c.name)
    tp.execute(f"Minkowski {c.name}")
    c.kcl.end_changes()

    return operator.region


def fix_width_minkowski_tiled(
    c: KCell,
    min_width: int,
    ref: kdb.LayerInfo | kdb.Region,
    n_threads: int | None = None,
    tile_size: tuple[float, float] | None = None,
    overlap: int = 1,
    smooth: int | None = None,
) -> kdb.Region:
    """Fix min space issues by using a dilation & erosion with a box.

    Args:
        c: Input cell
        min_width: Minimum width rule [dbu]
        ref: Input layer index or region
        n_threads: on how many threads to run the check simultaneously
        tile_size: tuple determining the size of each sub tile (in um), should be big
            compared to the violation size
        overlap: how many times bigger to make the tile border in relation to the
            violation size. Smaller than 1 can lead to errors
        smooth: Apply smoothening (simplifying) at the end if > 0

    Returns:
        kdb.Region: Region containing the fixes for the violations
    """
    tp = kdb.TilingProcessor()
    tp.frame = c.dbbox()  # type: ignore[misc]
    tp.dbu = c.kcl.dbu
    tp.threads = n_threads or config.n_threads

    min_tile_size_rec = 10 * min_width * tp.dbu

    if tile_size is None:
        tile_size = (min_tile_size_rec * 2, min_tile_size_rec * 2)

    tp.tile_border(min_width * overlap * tp.dbu, min_width * overlap * tp.dbu)

    tp.tile_size(*tile_size)
    if isinstance(ref, kdb.LayerInfo):
        tp.input("main_layer", c.kcl.layout, c.cell_index(), c.kcl.find_layer(ref))
    else:
        tp.input("main_layer", ref)

    operator = RegionOperator()
    tp.output("target", operator)
    if smooth is None:
        queue_str = (
            f"var tile_reg = (_tile & _frame).sized({min_width});"
            f"var shape = Box.new({min_width},{min_width});"
            "var reg = tile_reg - (tile_reg - main_layer).minkowski_sum(shape);"
            "reg = reg.minkowski_sum(shape); reg.merge();"
            "_output(target, reg & _tile, true);"
        )
    else:
        queue_str = (
            f"var tile_reg = (_tile & _frame).sized({min_width});"
            f"var shape = Box.new({min_width},{min_width});"
            "var reg = tile_reg - (tile_reg - main_layer).minkowski_sum(shape);"
            "reg = reg.minkowski_sum(shape); reg.merge();"
            f"reg.smooth({smooth});"
            "_output(target, reg & _tile, true);"
        )

    tp.queue(queue_str)
    logger.debug("String queued for {}:  {}", c.name, queue_str)

    c.kcl.start_changes()
    logger.info("Starting minkowski on {}", c.name)
    tp.execute(f"Minkowski {c.name}")
    c.kcl.end_changes()

    return operator.region


def fix_width_and_spacing_minkowski_tiled(
    c: KCell,
    min_space: int,
    min_width: int,
    ref: kdb.LayerInfo | kdb.Region,
    n_threads: int | None = None,
    tile_size: tuple[float, float] | None = None,
    overlap: int = 1,
    smooth: int | None = None,
) -> kdb.Region:
    """Fix min space and width issues by using a dilation & erosion with a box.

    The algorithm will dilate by min_space, erode by min_width + min_space, and
    finally dilate by min_width

    Args:
        c: Input cell
        min_space: Minimum space rule [dbu]
        min_width: Minimum width rule [dbu]
        ref: Input layer index or region
        n_threads: on how many threads to run the check simultaneously
        tile_size: tuple determining the size of each sub tile (in um), should be big
            compared to the violation size
        overlap: how many times bigger to make the tile border in relation to the
            violation size. Smaller than 1 can lead to errors (overlap*min_space)
        smooth: Apply smoothening (simplifying) at the end if > 0

    Returns:
        kdb.Region: Region containing the fixes for the violations
    """
    tp = kdb.TilingProcessor()
    tp.frame = c.dbbox()  # type: ignore[misc]
    tp.dbu = c.kcl.dbu
    tp.threads = n_threads or config.n_threads

    min_tile_size_rec = 10 * min_space * tp.dbu

    if tile_size is None:
        tile_size = (min_tile_size_rec * 2, min_tile_size_rec * 2)

    border = min_space * tp.dbu * overlap
    tp.tile_border(border, border)

    tp.tile_size(*tile_size)
    if isinstance(ref, kdb.LayerInfo):
        tp.input("main_layer", c.kcl.layout, c.cell_index(), c.kcl.find_layer(ref))
    else:
        tp.input("main_layer", ref)

    shrink = min_space + min_width

    operator = RegionOperator()
    tp.output("target", operator)
    if smooth is None:
        queue_str = (
            f"var tile_reg = (_tile & _frame).sized({min_space});"
            f"var space_shape = Box.new({min_space},{min_space});"
            f"var shrink_shape = Box.new({shrink},{shrink});"
            f"var width_shape = Box.new({min_width},{min_width});"
            "var reg = main_layer.minkowski_sum(space_shape); reg.merge();"
            "reg = tile_reg - (tile_reg - reg).minkowski_sum(shrink_shape);"
            "reg = reg.minkowski_sum(width_shape);"
            "_output(target, reg & _tile, true);"
        )
    else:
        queue_str = (
            f"var tile_reg = (_tile & _frame).sized({min_space});"
            f"var space_shape = Box.new({min_space},{min_space});"
            f"var shrink_shape = Box.new({shrink},{shrink});"
            f"var width_shape = Box.new({min_width},{min_width});"
            "var reg = main_layer.minkowski_sum(space_shape); reg.merge();"
            "reg = tile_reg - (tile_reg - reg).minkowski_sum(shrink_shape);"
            "reg = reg.minkowski_sum(width_shape);"
            f"reg.smooth({smooth});"
            "_output(target, reg & _tile, true);"
        )

    tp.queue(queue_str)
    logger.debug("String queued for {}:  {}", c.name, queue_str)

    c.kcl.start_changes()
    logger.info("Starting minkowski on {}", c.name)
    tp.execute(f"Minkowski {c.name}")
    c.kcl.end_changes()

    return operator.region


class RegionOperator(kdb.TileOutputReceiver):
    """Region collector. Just getst the tile and inserts it into the target cell."""

    def __init__(self) -> None:
        """Initialization."""
        self.region = kdb.Region()

    def put(
        self,
        ix: int,
        iy: int,
        tile: kdb.Box,
        region: kdb.Region,
        dbu: float,
        clip: bool,
    ) -> None:
        """Tiling Processor output call.

        Args:
            ix: x-axis index of tile.
            iy: y_axis index of tile.
            tile: The bounding box of the tile.
            region: The target object of the :py:class:~`klayout.db.TilingProcessor`
            dbu: dbu used by the processor.
            clip: Whether the target was clipped to the tile or not.
        """
        self.region.insert(region)



import kfactory as kf
import klayout.db as kdb
from conftest import Layers


def test_instance_xsize(LAYER: Layers) -> None:
    c = kf.KCell()
    ref = c << kf.cells.straight.straight(width=0.5, length=1, layer=LAYER.WG)
    assert ref.xsize


def test_instance_center(LAYER: Layers) -> None:
    c = kf.KCell()
    ref1 = c << kf.cells.straight.straight(width=0.5, length=1, layer=LAYER.WG)
    ref2 = c << kf.cells.straight.straight(width=0.5, length=1, layer=LAYER.WG)

    ref1.center = ref2.center
    ref2.center = ref1.center + kdb.Point(0, 1000).to_v()
    ref2.dmove((0, 10))


def test_instance_d_move(LAYER: Layers) -> None:
    c = kf.KCell()
    ref = c << kf.cells.straight.straight(width=0.5, length=1, layer=LAYER.WG)

    ref.dmovex(10)
    ref.dmovex(10.0)

    ref.dmovey(10)
    ref.dmovey(10.0)
    ref.dmovex(10).movey(10)
    ref.drotate(45).movey(10)

    ref.dxmin = 0
    ref.dxmax = 0
    ref.dymin = 0
    ref.dymax = 0

    ref.dmirror_y(0)
    ref.dmirror_x(0)


def test_mirror(LAYER: Layers) -> None:
    """Test arbitrary mirror."""
    c = kf.KCell()
    b = kf.cells.euler.bend_euler(width=1, radius=10, layer=LAYER.WG)

    c << b
    b2 = c << b
    disp = kdb.Trans(5000, 5000)
    # mp1 = kf.kdb.Point(-10000, 10000)
    mp1 = kf.kdb.Point(50000, 25000)
    mp2 = -mp1

    b2.mirror(disp * mp1, disp * mp2)

    c.shapes(c.kcl.find_layer(LAYER.WG)).insert(kf.kdb.Edge(mp1, mp2).transformed(disp))


def test_dmirror(LAYER: Layers) -> None:
    """Test arbitrary mirror."""
    c = kf.KCell()
    b = kf.cells.euler.bend_euler(width=1, radius=10, layer=LAYER.WG)

    c << b
    b2 = c << b
    disp = kdb.Trans(5000, 5000).to_dtype(c.kcl.dbu)
    # mp1 = kf.kdb.Point(-10000, 10000)
    mp1 = c.kcl.to_um(kf.kdb.Point(50000, 25000))
    mp2 = -mp1

    b2.dmirror(disp * mp1, disp * mp2)

    c.shapes(c.kcl.find_layer(LAYER.WG)).insert(
        kf.kdb.DEdge(mp1, mp2).transformed(disp)
    )



"""Virtual euler cell factories."""

from collections.abc import Callable
from typing import Any, Protocol

from ... import kdb
from ...conf import logger
from ...enclosure import LayerEnclosure
from ...factories.euler import euler_bend_points
from ...kcell import Info, KCLayout, MetaData, VKCell
from .utils import extrude_backbone


class BendEulerVKCell(Protocol):
    """Factory for virtual euler bends."""

    def __call__(
        self,
        width: float,
        radius: float,
        layer: kdb.LayerInfo,
        enclosure: LayerEnclosure | None = None,
        angle: float = 90,
        resolution: float = 150,
    ) -> VKCell:
        """Create a virtual euler bend.

        Args:
            width: Width of the core. [um]
            radius: Radius off the backbone. [um]
            layer: Layer index / LayerEnum of the core.
            enclosure: Slab/exclude definition. [dbu]
            angle: Angle of the bend.
            resolution: Angle resolution for the backbone.
        """
        ...


def virtual_bend_euler_factory(
    kcl: KCLayout,
    additional_info: Callable[
        ...,
        dict[str, MetaData],
    ]
    | dict[str, MetaData]
    | None = None,
    basename: str | None = None,
    **cell_kwargs: Any,
) -> BendEulerVKCell:
    """Returns a function generating virtual euler bends.

    Args:
        kcl: The KCLayout which will be owned
        additional_info: Add additional key/values to the
            [`VKCell.info`][kfactory.kcell.VKCell.info]. Can be a static dict
            mapping info name to info value. Or can a callable which takes the straight
            functions' parameters as kwargs and returns a dict with the mapping.
        basename: Overwrite the prefix of the resulting VKCell's name. By default
            the VKCell will be named 'virtual_bend_euler[...]'.
        cell_kwargs: Additional arguments passed as `@kcl.vcell(**cell_kwargs)`.
    """
    if callable(additional_info) and additional_info is not None:
        _additional_info_func: Callable[
            ...,
            dict[str, MetaData],
        ] = additional_info
        _additional_info: dict[str, MetaData] = {}
    else:

        def additional_info_func(
            **kwargs: Any,
        ) -> dict[str, MetaData]:
            return {}

        _additional_info_func = additional_info_func
        _additional_info = additional_info or {}

    @kcl.vcell(
        basename=basename,
        **cell_kwargs,
    )
    def bend_euler(
        width: float,
        radius: float,
        layer: kdb.LayerInfo,
        enclosure: LayerEnclosure | None = None,
        angle: float = 90,
        resolution: float = 150,
    ) -> VKCell:
        """Create a virtual euler bend.

        Args:
            width: Width of the core. [um]
            radius: Radius off the backbone. [um]
            layer: Layer index / LayerEnum of the core.
            enclosure: Slab/exclude definition. [dbu]
            angle: Angle of the bend.
            resolution: Angle resolution for the backbone.
        """
        c = VKCell()
        if angle < 0:
            logger.critical(
                f"Negative lengths are not allowed {angle} as ports"
                " will be inverted. Please use a positive number. Forcing positive"
                " lengths."
            )
            angle = -angle
        if width < 0:
            logger.critical(
                f"Negative widths are not allowed {width} as ports"
                " will be inverted. Please use a positive number. Forcing positive"
                " lengths."
            )
            width = -width
        dbu = c.kcl.dbu
        backbone = euler_bend_points(angle, radius=radius, resolution=resolution)

        extrude_backbone(
            c=c,
            backbone=backbone,
            width=width,
            layer=layer,
            enclosure=enclosure,
            start_angle=0,
            end_angle=angle,
            dbu=dbu,
        )
        _info: dict[str, MetaData] = {}
        _info.update(
            _additional_info_func(
                width=width,
                radius=radius,
                layer=layer,
                enclosure=enclosure,
                angle=angle,
            )
        )
        _info.update(_additional_info)
        c.info = Info(**_info)

        c.create_port(
            name="o1",
            layer=c.kcl.find_layer(layer),
            dwidth=width,
            dcplx_trans=kdb.DCplxTrans(1, 180, False, backbone[0].to_v()),
        )
        c.create_port(
            name="o2",
            dcplx_trans=kdb.DCplxTrans(1, angle, False, backbone[-1].to_v()),
            dwidth=width,
            layer=c.kcl.find_layer(layer),
        )
        return c

    return bend_euler



import kfactory as kf
import numpy as np
import warnings
from conftest import Layers


def bend_circular(
    width: int,
    radius: int,
    layer: kf.kdb.LayerInfo,
    enclosure: kf.enclosure.LayerEnclosure | None = None,
    angle: int = 90,
    angle_step: float = 1,
) -> kf.KCell:
    """Circular radius bend.

    Args:
        width: Width in database units
        radius: Radius in database units
        layer: Layer index of the target layer
        enclosure: :py:class:`kfactory.Enclosure` object to describe the claddings
    Returns:
        cell (KCell): Circular Bend KCell
    """
    c = kf.KCell()
    r = radius * c.kcl.dbu
    backbone = [
        kf.kdb.DPoint(x, y)
        for x, y in [
            [np.sin(_angle / 180 * np.pi) * r, (-np.cos(_angle / 180 * np.pi) + 1) * r]
            for _angle in np.linspace(
                0, angle, int(angle // angle_step + 0.5), endpoint=True
            )
        ]
    ]

    kf.enclosure.extrude_path(c, layer, backbone, width, enclosure, 0, angle)

    c.create_port(
        name="W0",
        trans=kf.kdb.Trans(2, False, 0, 0),
        width=width,
        layer=c.kcl.find_layer(layer),
    )

    match angle:
        case 90:
            c.create_port(
                name="N0",
                trans=kf.kdb.Trans(1, False, radius, radius),
                width=width,
                layer=c.kcl.find_layer(layer),
            )
        case 180:
            c.create_port(
                name="W1",
                trans=kf.kdb.Trans(0, False, 0, 2 * radius),
                width=width,
                layer=c.kcl.find_layer(layer),
            )

    return c


def dbend_circular(
    width: float,
    radius: float,
    layer: kf.kdb.LayerInfo,
    enclosure: kf.LayerEnclosure | None = None,
    angle: float = 90,
    angle_step: float = 1,
) -> kf.KCell:
    """Circular radius bend.

    Args:
        width: Width in database units
        radius: Radius in database units
        layer: Layer index of the target layer
        enclosure: :py:class:`kfactory.Enclosure` object to describe the claddings
    Returns:
        cell (KCell): Circular Bend KCell
    """
    c = kf.KCell()
    r = radius
    backbone = [
        kf.kdb.DPoint(x, y)
        for x, y in [
            [np.sin(_angle / 180 * np.pi) * r, (-np.cos(_angle / 180 * np.pi) + 1) * r]
            for _angle in np.linspace(
                0, angle, int(angle // angle_step + 0.5), endpoint=True
            )
        ]
    ]
    kf.enclosure.extrude_path(c, layer, backbone, width, enclosure, 0, angle)
    dp1 = kf.kcell.Port(
        dwidth=width,
        layer=c.kcl.find_layer(layer),
        name="W0",
        dcplx_trans=kf.kdb.DCplxTrans.R180,
    )
    warnings.filterwarnings("ignore")
    c.add_port(dp1)

    match angle:
        case 90:
            dp2 = kf.Port(
                name="N0",
                layer=c.kcl.find_layer(layer),
                dwidth=width,
                dcplx_trans=kf.kdb.DCplxTrans(1, 90, False, radius, radius),
            )
        case 180:
            dp2 = kf.Port(
                name="N0",
                layer=c.kcl.find_layer(layer),
                dwidth=width,
                dcplx_trans=kf.kdb.DTrans(1, 0, False, 0, 2 * radius),
            )
        case _:
            raise ValueError("only support 90/180° bends")
    c.add_port(dp2)
    warnings.filterwarnings("default")
    return c


def test_spiral(LAYER: Layers) -> None:
    c = kf.KCell()

    r1 = 1000
    r2 = 0

    p = kf.Port(
        name="start",
        trans=kf.kdb.Trans.R0,
        width=1000,
        layer=c.kcl.find_layer(LAYER.WG),
    )

    for _ in range(10):
        r = r1 + r2
        r2 = r1
        r1 = r
        b = c << bend_circular(width=1000, radius=r2, layer=LAYER.WG)
        b.connect("W0", p)
        p = b.ports["N0"]


def test_dspiral(LAYER: Layers) -> None:
    c = kf.KCell()

    r1 = 1
    r2 = 0

    p = kf.Port(
        name="start",
        dcplx_trans=kf.kdb.DCplxTrans.R0,
        dwidth=1,
        layer=c.kcl.find_layer(LAYER.WG),
    )

    kf.config.logfilter.level = kf.conf.LogLevel.ERROR

    for _ in range(10):
        r = r1 + r2
        r2 = r1
        r1 = r
        b = c << dbend_circular(width=1, radius=r2, layer=LAYER.WG)
        b.connect("W0", p)
        p = b.ports["N0"]



"""ill Utilities.

Filling empty regions in [KCells][kfactory.kcell.KCell] with filling cells.
"""

from collections.abc import Iterable

from .. import kdb
from ..conf import config, logger
from ..kcell import KCell, KCLayout

stop = False


class FillOperator(kdb.TileOutputReceiver):
    """Output Receiver of the TilingProcessor."""

    def __init__(
        self,
        kcl: KCLayout,
        top_cell: KCell,
        fill_cell_index: int,
        fc_bbox: kdb.Box,
        origin: kdb.Point,
        row_step: kdb.Vector,
        column_step: kdb.Vector,
        fill_margin: kdb.Vector = kdb.Vector(0, 0),
        remaining_polygons: kdb.Region | None = None,
        multi: bool = False,
    ) -> None:
        """Initialize the receiver."""
        self.kcl = kcl
        self.top_cell = top_cell
        self.fill_cell_index = fill_cell_index
        self.fc_bbox = fc_bbox
        self.row_step = row_step
        self.column_step = column_step
        self.fill_margin = fill_margin
        self.remaining_polygons = remaining_polygons
        self.multi = multi
        self.origin = origin

    def put(
        self,
        ix: int,
        iy: int,
        tile: kdb.Box,
        region: kdb.Region,
        dbu: float,
        clip: bool,
    ) -> None:
        """Called by the TilingProcessor."""
        if self.multi:
            self.top_cell.fill_region_multi(
                region=region,
                fill_cell_index=self.fill_cell_index,
                fc_bbox=self.fc_bbox,
                row_step=self.row_step,
                column_step=self.column_step,
                origin=self.origin,
                remaining_parts=None,
                fill_margin=kdb.Vector(
                    self.row_step.x - self.fc_bbox.width(),
                    self.column_step.y - self.fc_bbox.height(),
                ),
                remaining_polygons=None,
                glue_box=tile,
            )
        else:
            self.top_cell.fill_region(
                region=region,
                fill_cell_index=self.fill_cell_index,
                fc_bbox=self.fc_bbox,
                row_step=self.row_step,
                column_step=self.column_step,
                origin=self.origin,
                remaining_parts=None,
                fill_margin=self.fill_margin,
                remaining_polygons=None,
                glue_box=tile,
            )


def fill_tiled(
    c: KCell,
    fill_cell: KCell,
    fill_layers: Iterable[tuple[kdb.LayerInfo, int]] = [],
    fill_regions: Iterable[tuple[kdb.Region, int]] = [],
    exclude_layers: Iterable[tuple[kdb.LayerInfo, int]] = [],
    exclude_regions: Iterable[tuple[kdb.Region, int]] = [],
    n_threads: int | None = None,
    tile_size: tuple[float, float] | None = None,
    row_step: kdb.DVector | None = None,
    col_step: kdb.DVector | None = None,
    x_space: float = 0,
    y_space: float = 0,
    tile_border: tuple[float, float] = (20, 20),
    multi: bool = False,
) -> None:
    """Fill a [KCell][kfactory.kcell.KCell].

    Args:
        c: Target cell.
        fill_cell: The cell used as a cell to fill the regions.
        fill_layers: Tuples of layer and keepout w.r.t. the regions.
        fill_regions: Specific regions to fill. Also tuples like the layers.
        exclude_layers: Layers to ignore. Tuples like the fill layers
        exclude_regions: Specific regions to ignore. Tuples like the fill layers.
        n_threads: Max number of threads used. Defaults to number of cores of the
            machine.
        tile_size: Size of the tiles in um.
        row_step: DVector for steping to the next instance position in the row.
            x-coordinate must be >= 0.
        col_step: DVector for steping to the next instance position in the column.
            y-coordinate must be >= 0.
        x_space: Spacing between the fill cell bounding boxes in x-direction.
        y_space: Spacing between the fill cell bounding boxes in y-direction.
        tile_border: The tile border to consider for excludes
        multi: Use the region_fill_multi strategy instead of single fill.
    """
    if n_threads is None:
        n_threads = config.n_threads
    tp = kdb.TilingProcessor()
    dbb = c.dbbox()
    for r, ext in fill_regions:
        dbb += r.bbox().to_dtype(c.kcl.dbu).enlarged(ext)
    tp.frame = dbb  # type: ignore
    tp.dbu = c.kcl.dbu
    tp.threads = n_threads

    if tile_size is None:
        tile_size = (
            100 * (fill_cell.dbbox().width() + x_space),
            100 * (fill_cell.dbbox().height() + y_space),
        )
    tp.tile_size(*tile_size)
    tp.tile_border(*tile_border)

    layer_names: list[str] = []
    for _layer, _ in fill_layers:
        layer_name = (
            f"layer{_layer.name}"
            if _layer.is_named()
            else f"layer_{_layer.layer}_{_layer.datatype}"
        )
        tp.input(layer_name, c.kcl.layout, c.cell_index(), _layer)
        layer_names.append(layer_name)

    region_names: list[str] = []
    for i, (r, _) in enumerate(fill_regions):
        region_name = f"region{i}"
        tp.input(region_name, r)
        region_names.append(region_name)

    exlayer_names: list[str] = []
    for _layer, _ in exclude_layers:
        layer_name = (
            f"layer{_layer.name}"
            if _layer.is_named()
            else f"layer_{_layer.layer}_{_layer.datatype}"
        )
        tp.input(layer_name, c.kcl.layout, c.cell_index(), _layer)
        exlayer_names.append(layer_name)

    exregion_names: list[str] = []
    for i, (r, _) in enumerate(exclude_regions):
        region_name = f"region{i}"
        tp.input(region_name, r)
        exregion_names.append(region_name)

    if row_step is None:
        _row_step = kdb.Vector(fill_cell.bbox().width() + int(x_space / c.kcl.dbu), 0)
    else:
        _row_step = c.kcl.to_dbu(row_step)
    if col_step is None:
        _col_step = kdb.Vector(0, fill_cell.bbox().height() + int(y_space / c.kcl.dbu))
    else:
        _col_step = c.kcl.to_dbu(col_step)
    fc_bbox = fill_cell.bbox()
    tp.output(
        "to_fill",
        FillOperator(
            c.kcl,
            c,
            fill_cell.cell_index(),
            fc_bbox=fc_bbox,
            row_step=_row_step,
            column_step=_col_step,
            origin=c.bbox().p1,
        ),
    )

    if layer_names or region_names:
        exlayers = " + ".join(
            [
                layer_name + f".sized({c.kcl.to_dbu(size)})" if size else layer_name
                for layer_name, (_, size) in zip(exlayer_names, exclude_layers)
            ]
        )
        exregions = " + ".join(
            [
                region_name + f".sized({c.kcl.to_dbu(size)})" if size else region_name
                for region_name, (_, size) in zip(exregion_names, exclude_regions)
            ]
        )
        layers = " + ".join(
            [
                layer_name + f".sized({c.kcl.to_dbu(size)})" if size else layer_name
                for layer_name, (_, size) in zip(layer_names, fill_layers)
            ]
        )
        regions = " + ".join(
            [
                region_name + f".sized({c.kcl.to_dbu(size)})" if size else region_name
                for region_name, (_, size) in zip(region_names, fill_regions)
            ]
        )

        if exlayer_names or exregion_names:
            queue_str = (
                "var fill= "
                + (
                    " + ".join([layers, regions])
                    if regions and layers
                    else regions + layers
                )
                + "; var exclude = "
                + (
                    " + ".join([exlayers, exregions])
                    if exregions and exlayers
                    else exregions + exlayers
                )
                + "; var fill_region = _tile.minkowski_sum(Box.new("
                f"0,0,{fc_bbox.width() - 1},{fc_bbox.height() - 1}))"
                " & _frame & fill - exclude; _output(to_fill, fill_region)"
            )
        else:
            queue_str = (
                "var fill= "
                + (
                    " + ".join([layers, regions])
                    if regions and layers
                    else regions + layers
                )
                + "; var fill_region = _tile.minkowski_sum(Box.new("
                f"0,0,{fc_bbox.width() - 1},{fc_bbox.height() - 1}))"
                " & _frame & fill;"
                " _output(to_fill, fill_region)"
            )
        tp.queue(queue_str)
        c.kcl.start_changes()
        try:
            logger.debug(
                "Filling {} with {}", c.kcl.future_cell_name or c.name, fill_cell.name
            )
            logger.debug("fill string: '{}'", queue_str)
            tp.execute(f"Fill {c.name}")
            logger.info("done with filling {}", c.name)
        finally:
            c.kcl.end_changes()



"""Utilities to provide geometrical, fill and DRC violation help.

[fill_tiled][kfactory.utils.fill_tiled] provides a filling algorithm that can use
the `klayout.db.TilingProcessor` to calculate the regions to fill.

[fix_spacing][kfactory.utils.violations.fix_spacing_tiled] uses a region space check to
calculate areas that violate min space violations.
"""

# from .enclosure import (
#     Direction,
#     KCellEnclosure,
#     LayerEnclosure,
#     clean_points,
#     extrude_path,
#     extrude_path_dynamic,
# )

from .fill import fill_tiled
from .violations import fix_spacing_minkowski_tiled, fix_spacing_tiled

__all__ = [
    "fill_tiled",
    "fix_spacing_tiled",
    "fix_spacing_minkowski_tiled",
]



import pytest
import kfactory as kf
from collections.abc import Callable
from tempfile import NamedTemporaryFile
from conftest import Layers


def test_enclosure_name(straight_factory_dbu: Callable[..., kf.KCell]) -> None:
    wg = straight_factory_dbu(width=1000, length=10000)
    assert wg.name == "straight_W1000_L10000_LWG_EWGSTD"


def test_circular_snapping(LAYER: Layers) -> None:
    b = kf.cells.circular.bend_circular(width=1, radius=10, layer=LAYER.WG, angle=90)
    assert b.ports["o2"].dcplx_trans.disp == kf.kcl.to_um(b.ports["o2"].trans.disp)


def test_euler_snapping(LAYER: Layers) -> None:
    b = kf.cells.euler.bend_euler(width=1, radius=10, layer=LAYER.WG, angle=90)
    assert b.ports["o2"].dcplx_trans.disp == kf.kcl.to_um(b.ports["o2"].trans.disp)


@kf.cell
def unnamed_cell(name: str = "a") -> kf.KCell:
    c = kf.kcl.kcell(name)
    return c


def test_unnamed_cell() -> None:
    c1 = unnamed_cell("test")
    c2 = unnamed_cell("test")
    assert c1 is c2


@kf.cell
def nested_list_dict(
    arg1: dict[str, list[dict[str, str | int] | int] | int],
) -> kf.KCell:
    c = kf.kcl.kcell()
    return c


def test_nested_dict_list() -> None:
    dl: dict[str, list[dict[str, str | int] | int] | int] = {
        "a": 5,
        "b": [5, {"c": "d", "e": 6}],
    }
    c = nested_list_dict(dl)
    assert dl == c.settings["arg1"]
    assert dl is not c.settings["arg1"]


def test_no_snap(LAYER: Layers) -> None:
    c = kf.KCell()

    c.create_port(
        dwidth=1,
        dcplx_trans=kf.kdb.DCplxTrans(1, 90, False, 0.0005, 0),
        layer=c.kcl.find_layer(LAYER.WG),
    )

    p = c.ports[0]

    assert p.dcplx_trans.disp != c.kcl.to_um(p.trans.disp)


def test_namecollision(LAYER: Layers) -> None:
    b1 = kf.cells.circular.bend_circular(width=1, radius=10.5, layer=LAYER.WG)
    b2 = kf.cells.circular.bend_circular(width=1, radius=10.5000005, layer=LAYER.WG)

    assert b1.name != b2.name


def test_nested_dic() -> None:
    @kf.kcl.cell
    def recursive_dict_cell(d: dict[str, dict[str, str] | str]) -> kf.KCell:
        c = kf.KCell()
        return c

    recursive_dict_cell({"test": {"test2": "test3"}, "test4": "test5"})


def test_ports_cell(LAYER: Layers) -> None:
    c = kf.KCell()
    c.create_port(
        name="o1",
        dwidth=1,
        dcplx_trans=kf.kdb.DCplxTrans(1, 90, False, 0.0005, 0),
        layer=c.kcl.find_layer(LAYER.WG),
    )
    assert c["o1"]
    assert "o1" in c.ports


def test_ports_instance(LAYER: Layers) -> None:
    c = kf.KCell()
    c.create_port(
        name="o1",
        dwidth=1,
        dcplx_trans=kf.kdb.DCplxTrans(1, 90, False, 0.0005, 0),
        layer=c.kcl.find_layer(LAYER.WG),
    )
    c2 = kf.KCell()
    ref = c2 << c
    assert c["o1"]
    assert "o1" in c.ports
    assert ref["o1"]
    assert "o1" in ref.ports


def test_getter(LAYER: Layers) -> None:
    c = kf.KCell()
    c << kf.cells.straight.straight(width=1, length=10, layer=LAYER.WG)
    assert c.y == 0
    assert c.dy == 0


def test_array(straight: kf.KCell) -> None:
    c = kf.KCell()
    wg_array = c.create_inst(
        straight, a=kf.kdb.Vector(15_000, 0), b=kf.kdb.Vector(0, 3_000), na=3, nb=5
    )
    for b in range(5):
        for a in range(3):
            wg_array["o1", a, b]
            wg_array["o1", a, b]


def test_array_indexerror(straight: kf.KCell) -> None:
    c = kf.KCell()
    wg_array = c.create_inst(
        straight, a=kf.kdb.Vector(15_000, 0), b=kf.kdb.Vector(0, 3_000), na=3, nb=5
    )
    regex = kf.config.logfilter.regex
    kf.config.logfilter.regex = r"^An error has been caught in function '__getitem__'"
    with pytest.raises(IndexError):
        wg_array["o1", 3, 5]
        wg_array["o1", 3, 5]
    kf.config.logfilter.regex = regex


def test_invalid_array(monkeypatch: pytest.MonkeyPatch, straight: kf.KCell) -> None:
    c = kf.KCell()
    wg = c.create_inst(straight)
    regex = kf.config.logfilter.regex
    kf.config.logfilter.regex = r"^An error has been caught in function '__getitem__'"
    with pytest.raises(KeyError):
        for b in range(1):
            for a in range(1):
                wg["o1", a, b]
                wg["o1", a, b]
    kf.config.logfilter.regex = regex


def test_cell_decorator_error(LAYER: Layers) -> None:
    kcl2 = kf.KCLayout("decorator_test")

    @kf.kcl.cell
    def wrong_cell() -> kf.KCell:
        c = kcl2.kcell("wrong_test")
        return c

    regex = kf.config.logfilter.regex
    kf.config.logfilter.regex = (
        r"^An error has been caught in function 'wrapper_autocell'"
    )
    with pytest.raises(ValueError):
        wrong_cell()
    kf.config.logfilter.regex = regex


def test_info(LAYER: Layers) -> None:
    @kf.kcl.cell(info={"test": 42})
    def test_info_cell(test: int) -> kf.KCell:
        return kf.kcl.kcell()

    c = test_info_cell(42)
    assert c.info["test"] == 42


def test_flatten(LAYER: Layers) -> None:
    c = kf.KCell()
    _ = c << kf.cells.straight.straight(width=1, length=10, layer=LAYER.WG)
    assert len(c.insts) == 1, "c.insts should have 1 inst after adding a cell"
    c.flatten()
    assert len(c.insts) == 0, "c.insts should have 0 insts after flatten()"


def test_size_info(LAYER: Layers) -> None:
    c = kf.KCell()
    ref = c << kf.cells.straight.straight(width=1, length=10, layer=LAYER.WG)
    assert ref.size_info.ne[0] == 10000
    assert ref.dsize_info.ne[0] == 10


def test_overwrite(LAYER: Layers) -> None:
    kcl = kf.KCLayout("CELL_OVERWRITE")

    @kcl.cell
    def test_overwrite_cell() -> kf.KCell:
        print("overwrite1")
        c = kcl.kcell()
        return c

    c1 = test_overwrite_cell()

    @kcl.cell(overwrite_existing=True)  # type: ignore[no-redef]
    def test_overwrite_cell() -> kf.KCell:
        print("overwrite2")
        c = kcl.kcell()
        return c

    c2 = test_overwrite_cell()

    assert c2 is not c1
    assert c1._destroyed()


def test_layout_cache(LAYER: Layers) -> None:
    kcl_write = kf.KCLayout("TEST_LAYOUT_CACHE_WRITE")
    kcl_read = kf.KCLayout("TEST_LAYOUT_CACHE_READ")

    @kcl_write.cell(basename="straight")
    def write_straight() -> kf.KCell:
        c = kcl_write.kcell()
        c.shapes(kcl_write.layer(1, 0)).insert(kf.kdb.Box(10_000, 1000))
        return c

    s_write = write_straight()
    tf = NamedTemporaryFile(suffix=".gds.gz")
    kcl_write.write(tf.name)
    kcl_read.read(tf.name)

    @kcl_read.cell(basename="straight", layout_cache=True)
    def read_straight() -> kf.KCell:
        c = kcl_read.kcell()
        c.shapes(kcl_read.layer(1, 0)).insert(kf.kdb.Box(5000, 1000))
        return c

    s_read = read_straight()
    assert s_write.bbox() == s_read.bbox()


def test_check_ports(LAYER: Layers) -> None:
    kcl = kf.KCLayout("CHECK_PORTS", infos=Layers)
    kcl.layers = kcl.layerenum_from_dict(layers=LAYER)

    @kcl.cell
    def test_multi_ports() -> kf.KCell:
        c = kcl.kcell()
        c.create_port(
            name="a", trans=kf.kdb.Trans.R0, width=1000, layer=kcl.find_layer(1, 0)
        )
        c.create_port(
            name="a", trans=kf.kdb.Trans.R180, width=1000, layer=kcl.find_layer(1, 0)
        )
        return c

    regex = kf.config.logfilter.regex
    kf.config.logfilter.regex = "^An error has been caught in function 'wrapper_autocell', process 'MainProcess'"

    with pytest.raises(ValueError):
        test_multi_ports()

    kf.config.logfilter.regex = regex



"""Interactivate jupyter widget."""

from __future__ import annotations

try:
    from pathlib import Path
    from typing import Any

    from ipyevents import Event  # type: ignore[import-untyped,unused-ignore]
    from IPython.display import Image as IPImage
    from IPython.display import display
    from ipytree import Node, Tree  # type: ignore[import-untyped,unused-ignore]
    from ipywidgets import (  # type: ignore[import-untyped,unused-ignore]
        Accordion,
        AppLayout,
        Box,
        Button,
        HBox,
        Image,
        Label,
        Layout,
        Output,
        RadioButtons,
        Tab,
        ToggleButtons,
        VBox,
    )

    from .. import kdb, lay
    from ..conf import config, logger
    from ..kcell import KCell
    from typing import Literal

except ImportError as e:
    print("You need install jupyter notebook plugin with `pip install kfactory[ipy]`")
    raise e

__all__ = ["display_kcell"]


def display_kcell(
    kc: KCell,
    lyrdb: Path | str | None = None,
    display_type: Literal["image", "widget"] | None = None,
) -> None:
    """Display a KCell in a jupyter widget or an image."""
    cell_dup = kc.kcl[kc.name].dup()
    cell_dup.insert_vinsts()
    display_type = display_type or config.display_type
    match display_type:
        case "widget":
            lw = LayoutWidget(cell=cell_dup, layer_properties=lyrdb)
            display(lw.widget)  # type: ignore[no-untyped-call,unused-ignore]
        case "image":
            lipi = LayoutIPImage(cell=cell_dup, layer_properties=lyrdb)
            display(lipi.image)  # type: ignore[no-untyped-call,unused-ignore]
    cell_dup.delete()


widgets: list[LayoutImage | LayoutIPImage] = []


class LayoutImage:
    def __init__(
        self,
        cell: KCell,
        layer_properties: str | None = None,
    ):
        self.layout_view = lay.LayoutView()
        self.layout_view.show_layout(cell.kcl.layout, False)
        self.layer_properties: Path | None = None
        if layer_properties is not None:
            self.layer_properties = Path(layer_properties)
            if self.layer_properties.exists() and self.layer_properties.is_file():
                self.layer_properties = self.layer_properties
                self.layout_view.load_layer_props(str(self.layer_properties))
        self.show_cell(cell._kdb_cell)
        png_data = self.layout_view.get_screenshot_pixels().to_png_data()

        self.image = Image(value=png_data, format="png")
        widgets.append(self)

    def show_cell(self, cell: kdb.Cell) -> None:
        self.layout_view.active_cellview().cell = cell
        self.layout_view.max_hier()
        self.layout_view.resize(800, 600)
        self.layout_view.add_missing_layers()
        self.layout_view.zoom_fit()


class LayoutIPImage:
    def __init__(
        self,
        cell: KCell,
        layer_properties: Path | str | None = None,
    ):
        self.layout_view = lay.LayoutView()
        self.layout_view.show_layout(cell.kcl.layout, False)
        self.layer_properties: Path | None = None
        if layer_properties is not None:
            self.layer_properties = Path(layer_properties)
            if self.layer_properties.exists() and self.layer_properties.is_file():
                self.layer_properties = self.layer_properties
                self.layout_view.load_layer_props(str(self.layer_properties))
        self.show_cell(cell._kdb_cell)
        png_data = self.layout_view.get_screenshot_pixels().to_png_data()
        self.image = IPImage(  # type: ignore[no-untyped-call,unused-ignore]
            data=png_data, format="png", embed=True, width=800, height=600
        )
        widgets.append(self)

    def show_cell(self, cell: kdb.Cell) -> None:
        self.layout_view.active_cellview().cell = cell
        self.layout_view.max_hier()
        self.layout_view.resize(800, 600)
        self.layout_view.add_missing_layers()
        self.layout_view.zoom_fit()


class LayoutWidget:
    def __init__(
        self,
        cell: KCell,
        layer_properties: Path | str | None = None,
        hide_unused_layers: bool = True,
        with_layer_selector: bool = True,
    ):
        self.debug = Output()

        self.hide_unused_layers = hide_unused_layers

        self.layout_view = lay.LayoutView()
        self.layout_view.show_layout(cell.kcl.layout, False)
        self.layer_properties: Path | None = None
        if layer_properties is not None:
            self.layer_properties = Path(layer_properties)
            if self.layer_properties.exists() and self.layer_properties.is_file():
                self.layer_properties = self.layer_properties
                self.layout_view.load_layer_props(str(self.layer_properties))
        self.show_cell(cell._kdb_cell)
        png_data = self.layout_view.get_screenshot_pixels().to_png_data()

        self.image = Image(value=png_data, format="png")
        self.refresh()
        scroll_event = Event(source=self.image, watched_events=["wheel"], wait=10)
        scroll_event.on_dom_event(self.on_scroll)

        enter_event = Event(source=self.image, watched_events=["mouseenter"])
        leave_event = Event(source=self.image, watched_events=["mouseleave"])
        enter_event.on_dom_event(self.on_mouse_enter)
        leave_event.on_dom_event(self.on_mouse_leave)

        self.layout_view.on_image_updated_event = self.refresh  # type: ignore[assignment]  # noqa: E501

        mouse_event = Event(
            source=self.image,
            watched_events=[
                "mousedown",
                "mouseup",
                "mousemove",
                "contextmenu",
            ],
            wait=10,
            throttle_or_debounce="debounce",
            prevent_default_action=True,
        )
        mouse_event.on_dom_event(self.on_mouse)

        max_height = self.layout_view.viewport_height()
        if with_layer_selector:
            selector_tabs = self.build_selector(max_height=max_height)
        else:
            selector_tabs = None

        mode_selector = self.build_modes(max_height)

        def switch_mode(buttons: ToggleButtons) -> None:
            self.layout_view.switch_mode(buttons.value)
            self.refresh()

        self.widget = AppLayout(
            left_sidebar=mode_selector,
            center=self.image,
            right_sidebar=selector_tabs,
            connect_items="top",
            justify_items="center",
            footer=self.debug,
            # footer=mode_selector,
            pane_weights=[1, 3, 1],
        )

    def show_cell(self, cell: kdb.Cell) -> None:
        self.layout_view.active_cellview().cell_index = cell.cell_index()
        self.layout_view.max_hier()
        self.layout_view.resize(800, 600)
        self.layout_view.add_missing_layers()
        self.layout_view.zoom_fit()

    def button_toggle(self, button: Button) -> None:
        button.style.button_color = (
            "transparent"
            if (button.style.button_color == button.default_color)
            else button.default_color
        )

        logger.info("button toggle")
        for props in self.layout_view.each_layer():
            if props == button.layer_props:
                props.visible = not props.visible
                props.name = button.name
                button.layer_props = props
                break
        self.refresh()

    def build_layer_toggle(self, prop_iter: lay.LayerPropertiesIterator) -> HBox | None:
        props = prop_iter.current()
        layer_color = f"#{props.eff_fill_color():06x}"
        Layout(
            width="5px",
            height="20px",
            border=f"solid 2px {layer_color}",
            display="block",
        )

        # prop_iter.current().visible = False

        image = Image(
            value=self.layout_view.icon_for_layer(prop_iter, 50, 25, 1).to_png_data(),
            format="png",
            width=50,
            height=25,
        )

        image_event = Event(source=image, watched_events=["click"])
        _prop = prop_iter.dup()

        def on_layer_click(event: Event) -> None:
            _prop.current().visible = not _prop.current().visible
            self.layout_view.timer()
            image.value = self.layout_view.icon_for_layer(
                _prop, 50, 25, 1
            ).to_png_data()

            self.refresh()

        image_event.on_dom_event(on_layer_click)

        if props.has_children():
            prop_iter = prop_iter.down_first_child()
            n_children = prop_iter.num_siblings()
            children = []
            for _i in range(n_children):
                prop_iter = prop_iter.next()
                _layer_toggle = self.build_layer_toggle(prop_iter)
                if _layer_toggle is not None:
                    children.append(_layer_toggle)

            if not children:
                return None
            layer_label = Accordion([VBox(children)], titles=(props.name,))

        else:
            cell = self.layout_view.active_cellview().cell
            if cell.bbox(prop_iter.current().layer_index()).empty():
                return None
            layer_label = (
                Label(props.name)
                if props.name
                else Label(f"{props.source_layer}/{props.source_datatype}")
            )

        return HBox([Box([image]), layer_label])

    def build_modes(self, max_height: float) -> VBox:
        def clear(event: Event) -> None:
            self.layout_view.clear_annotations()
            self.refresh()

        clear_button = Button(description="clear annotations")
        clear_button.on_click(clear)

        def zoom_fit(even: Event) -> None:
            self.layout_view.zoom_fit()
            self.refresh()

        zoom_button = Button(description="zoom fit")
        zoom_button.on_click(zoom_fit)

        mode_label = Label("mode selection:")

        modes = self.layout_view.mode_names()
        tb = ToggleButtons(options=modes)

        def switch_mode(mode: dict[str, Any]) -> None:
            self.layout_view.switch_mode(mode["new"])
            self.refresh()

        tb.observe(switch_mode, "value")
        return VBox(children=(clear_button, zoom_button, mode_label, tb))

    def build_cell_selector(self, cell: kdb.Cell) -> Node:
        child_cells = [
            self.build_cell_selector(
                self.layout_view.active_cellview().layout().cell(_cell)
            )
            for _cell in cell.each_child_cell()
        ]

        node = Node(cell.name, show_icon=False)
        node.observe(self.on_select_cell, "selected")

        for cc in child_cells:
            node.add_node(cc)

        return node

    def build_selector(self, max_height: float) -> Tab:
        """Builds a widget for toggling layer displays.

        Args:
            max_height: Maximum height to set for the widget (likely the height of the
                pixel buffer).
        """
        all_boxes = []

        prop_iter = self.layout_view.begin_layers()
        while not prop_iter.at_end():
            if layer_toggle := self.build_layer_toggle(prop_iter):
                all_boxes.append(layer_toggle)
            prop_iter.next()

        layout = Layout(
            max_height=f"{max_height}px", overflow_y="auto", display="block"
        )
        selector = VBox(all_boxes, layout=layout)

        cells: list[RadioButtons | Accordion] = [
            self.build_cell_selector(cell)
            for cell in self.layout_view.active_cellview().layout().top_cells()
        ]
        tree = Tree(cells, multiple_selection=False, stripes=True)

        # For when tabs are implemented
        selector_tabs = Tab([selector, tree])
        selector_tabs.set_title(0, "Layers")
        selector_tabs.set_title(1, "Cells")
        # selector_tabs.titles = ("Layers",)

        return selector_tabs

    def load_layout(self, filepath: str, layer_properties: str | None) -> None:
        """Loads a GDS layout.

        Args:
            filepath: path for the GDS layout.
            layer_properties: Optional path for the layer_properties klayout file (lyp).
        """
        self.layout_view.load_layout(filepath)
        self.layout_view.max_hier()
        if layer_properties:
            self.layout_view.load_layer_props(layer_properties)

    def refresh(self) -> None:
        self.layout_view.timer()
        png_data = self.layout_view.get_screenshot_pixels().to_png_data()
        self.image.value = png_data
        self.layout_view.timer()

    def _get_modifier_buttons(self, event: Event) -> int:
        shift = event["shiftKey"]
        alt = event["altKey"]
        ctrl = event["ctrlKey"]
        # meta = event["metaKey"]

        mouse_buttons = event["buttons"]

        buttons = 0
        if shift:
            buttons |= lay.ButtonState.ShiftKey
        if alt:
            buttons |= lay.ButtonState.AltKey
        if ctrl:
            buttons |= lay.ButtonState.ControlKey

        if mouse_buttons & 1:
            buttons |= lay.ButtonState.LeftButton
        if mouse_buttons & 2:
            buttons |= lay.ButtonState.RightButton
        if mouse_buttons & 4:
            buttons |= lay.ButtonState.MidButton

        return buttons

    def on_select_cell(self, event: Event) -> None:
        self.show_cell(
            self.layout_view.active_cellview().layout().cell(event["owner"].name)
        )

        self.layout_view.viewport_height()

        all_boxes = []
        prop_iter = self.layout_view.begin_layers()
        while not prop_iter.at_end():
            if layer_toggle := self.build_layer_toggle(prop_iter):
                all_boxes.append(layer_toggle)
            prop_iter.next()

        tabs = self.widget.right_sidebar
        vbox = tabs.children[0]
        vbox.children = all_boxes
        self.refresh()

    def on_scroll(self, event: Event) -> None:
        self.layout_view.timer()
        delta = int(event["deltaY"])
        x = event["relativeX"]
        y = event["relativeY"]
        buttons = self._get_modifier_buttons(event)
        self.layout_view.send_wheel_event(-delta, False, kdb.DPoint(x, y), buttons)
        self.refresh()

    def on_mouse(self, event: Event) -> None:
        self.layout_view.timer()
        x = event["relativeX"]
        y = event["relativeY"]
        buttons = self._get_modifier_buttons(event)

        match event["event"]:
            case "mousedown":
                self.layout_view.send_mouse_press_event(
                    kdb.DPoint(float(x), float(y)), buttons
                )
            case "mouseup":
                self.layout_view.send_mouse_release_event(
                    kdb.DPoint(float(x), float(y)), buttons
                )
            case "mousemove":
                self.layout_view.send_mouse_move_event(
                    kdb.DPoint(float(x), float(y)), buttons
                )
        self.refresh()
        self.layout_view.timer()

    def on_mouse_enter(self, event: Event) -> None:
        self.layout_view.timer()
        self.layout_view.send_enter_event()
        self.refresh()

    def on_mouse_leave(self, event: Event) -> None:
        self.layout_view.timer()
        self.layout_view.send_leave_event()
        self.refresh()



import pathlib
import pytest

import kfactory as kf

from functools import partial
from kfactory.conf import logger
from conftest import Layers


class GeometryDifference(ValueError):
    """Exception for Geometric differences."""

    pass


# class LAYER(kf.LayerEnum):  # type: ignore[unused-ignore, misc]
#     kcl = kf.constant(kf.kcl)
#     WG = (1, 0)
#     WGCLAD = (111, 0)


wg_enc = kf.LayerEnclosure(name="WGSTD", sections=[(Layers().WGCLAD, 0, 2000)])


def straight(LAYER: Layers) -> kf.KCell:
    return kf.cells.straight.straight(
        width=0.5, length=1, layer=LAYER.WG, enclosure=wg_enc
    )


def bend90(LAYER: Layers) -> kf.KCell:
    return kf.cells.circular.bend_circular(
        width=1, radius=10, layer=LAYER.WG, enclosure=wg_enc, angle=90
    )


def bend180(LAYER: Layers) -> kf.KCell:
    return kf.cells.circular.bend_circular(
        width=1, radius=10, layer=LAYER.WG, enclosure=wg_enc, angle=180
    )


def bend90_euler(LAYER: Layers) -> kf.KCell:
    return kf.cells.euler.bend_euler(
        width=1, radius=10, layer=LAYER.WG, enclosure=wg_enc, angle=90
    )


def bend180_euler(LAYER: Layers) -> kf.KCell:
    return kf.cells.euler.bend_euler(
        width=1, radius=10, layer=LAYER.WG, enclosure=wg_enc, angle=180
    )


def taper(LAYER: Layers) -> kf.KCell:
    c = kf.cells.taper.taper(
        width1=0.5,
        width2=1,
        length=10,
        layer=LAYER.WG,
        enclosure=wg_enc,
    )
    c = c.dup()
    c.name = "taper"
    return c


cells = dict(
    bend90=bend90,
    bend180=bend180,
    bend180_euler=bend180_euler,
    bend90_euler=bend90_euler,
    taper=taper,
    straight=straight,
)

cell_names = set(cells.keys())


@pytest.fixture(params=cell_names, scope="function")
def cell_name(request: pytest.FixtureRequest) -> str:
    """Returns cell name."""
    return request.param  # type: ignore[no-any-return]


def test_cells(cell_name: str, LAYER: Layers) -> None:
    """Ensure cells have the same geometry as their golden references."""
    gds_ref = pathlib.Path(__file__).parent / "test_data" / "ref"
    cell = cells[cell_name](LAYER)
    ref_file = gds_ref / f"{cell.name}.gds"
    run_cell = cell
    if not ref_file.exists():
        gds_ref.mkdir(parents=True, exist_ok=True)
        run_cell.write(str(ref_file))
        raise FileNotFoundError(f"GDS file not found. Saving it to {ref_file}")
    kcl_ref = kf.KCLayout("TEST", infos=Layers)
    kcl_ref.read(gds_ref / f"{cell.name}.gds", meta_format="v2", register_cells=True)
    ref_cell = kcl_ref[kcl_ref.top_cell().name]

    for layerinfo in kcl_ref.layout.layer_infos():
        layer = kcl_ref.layer(layerinfo)
        region_run = kf.kdb.Region(run_cell.begin_shapes_rec(layer))
        region_ref = kf.kdb.Region(ref_cell.begin_shapes_rec(layer))

        region_diff = region_run - region_ref

        if not region_diff.is_empty():
            layer_tuple = kcl_ref.layout.layer_infos()[layer]
            region_xor = region_ref ^ region_run
            c = kf.KCell(f"{cell.name}_diffs")
            c_run = kf.KCell(f"{cell.name}_new")
            c_ref = kf.KCell(f"{cell.name}_old")
            c_xor = kf.KCell(f"{cell.name}_xor")
            c_run.shapes(layer).insert(region_run)
            c_ref.shapes(layer).insert(region_ref)
            c_xor.shapes(layer).insert(region_xor)
            c << c_run
            c << c_ref
            c << c_xor

            kf.logger.critical(f"Differences found in {cell!r} on layer {layer_tuple}")
            val = input("Save current GDS as new reference (Y)? [Y/n]")
            if not val.upper().startswith("N"):
                logger.info(f"replacing file {str(ref_file)!r}")
                run_cell.write(ref_file.name)

            raise GeometryDifference(
                f"Differences found in {cell!r} on layer {layer_tuple}"
            )


def test_additional_info(LAYER: Layers, wg_enc: kf.LayerEnclosure) -> None:
    test_bend_euler = partial(
        kf.factories.euler.bend_euler_factory(
            kcl=kf.kcl,
            additional_info={"creation_time": "2023-02-12Z23:00:00"},
            overwrite_existing=True,
        ),
        layer=LAYER.WG,
        radius=10,
    )

    bend = test_bend_euler(width=1)

    assert bend._locked is True
    assert bend.info.creation_time == "2023-02-12Z23:00:00"  # type: ignore[attr-defined, unused-ignore]

    bend.delete()



"""CrossSections for KFactory."""

from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING

from pydantic import BaseModel, model_validator
from typing_extensions import Self

from . import kdb
from .enclosure import DLayerEnclosure, LayerEnclosure

if TYPE_CHECKING:
    from .kcell import KCLayout


class SymmetricalCrossSection(BaseModel, frozen=True):
    """CrossSection which is symmetrical to its main_layer/width."""

    width: int
    enclosure: LayerEnclosure
    name: str

    def __init__(
        self, width: int, enclosure: LayerEnclosure, name: str | None = None
    ) -> None:
        """Initialized the CrossSection."""
        super().__init__(
            width=width, enclosure=enclosure, name=name or f"{enclosure.name}_{width}"
        )

    @model_validator(mode="after")
    def _validate_enclosure_main_layer(self) -> Self:
        if self.enclosure.main_layer is None:
            raise ValueError("Enclosures of cross sections must have a main layer.")
        if (self.width // 2) * 2 != self.width:
            raise ValueError(
                "Width of symmetrical cross sections must have be a multiple of 2. "
                "This could cause cross sections and extrusions to become unsymmetrical"
                " otherwise."
            )
        return self

    @cached_property
    def main_layer(self) -> kdb.LayerInfo:
        """Main Layer of the enclosure and cross section."""
        return self.enclosure.main_layer  # type: ignore[return-value]

    def to_dtype(self, kcl: KCLayout) -> DSymmetricalCrossSection:
        """Convert to a um based CrossSection."""
        return DSymmetricalCrossSection(
            width=kcl.to_um(self.width),
            enclosure=self.enclosure.to_dtype(kcl),
            name=self.name,
        )


class DSymmetricalCrossSection(BaseModel):
    """um based CrossSection."""

    width: float
    enclosure: DLayerEnclosure
    name: str | None = None

    def to_itype(self, kcl: KCLayout) -> SymmetricalCrossSection:
        """Convert to a dbu based CrossSection."""
        return SymmetricalCrossSection(
            width=kcl.to_dbu(self.width),
            enclosure=kcl.get_enclosure(self.enclosure.to_itype(kcl)),
            name=self.name,
        )



"""Straight waveguide in dbu.

A waveguide is a rectangle of material with excludes and/or slab around it::

    ┌──────────────────────────────┐
    │         Slab/Exclude         │
    ├──────────────────────────────┤
    │                              │
    │             Core             │
    │                              │
    ├──────────────────────────────┤
    │         Slab/Exclude         │
    └──────────────────────────────┘

The slabs and excludes can be given in the form of an
[Enclosure][kfactory.enclosure.LayerEnclosure].
"""

from collections.abc import Callable
from typing import Any, Protocol

from .. import kdb, kf_types
from ..conf import logger
from ..enclosure import LayerEnclosure
from ..kcell import Info, KCell, KCLayout, MetaData

__all__ = ["straight_dbu_factory"]


class StraightKCellFactory(Protocol):
    def __call__(
        self,
        width: kf_types.dbu,
        length: kf_types.dbu,
        layer: kdb.LayerInfo,
        enclosure: LayerEnclosure | None = None,
    ) -> KCell:
        """Waveguide defined in dbu.

            ┌──────────────────────────────┐
            │         Slab/Exclude         │
            ├──────────────────────────────┤
            │                              │
            │             Core             │
            │                              │
            ├──────────────────────────────┤
            │         Slab/Exclude         │
            └──────────────────────────────┘
        Args:
            width: Waveguide width. [dbu]
            length: Waveguide length. [dbu]
            layer: Main layer of the waveguide.
            enclosure: Definition of slab/excludes. [dbu]
        """
        ...


def straight_dbu_factory(
    kcl: KCLayout,
    additional_info: Callable[
        ...,
        dict[str, MetaData],
    ]
    | dict[str, MetaData]
    | None = None,
    basename: str | None = None,
    **cell_kwargs: Any,
) -> StraightKCellFactory:
    """Returns a function generating straights [dbu].

        ┌──────────────────────────────┐
        │         Slab/Exclude         │
        ├──────────────────────────────┤
        │                              │
        │             Core             │
        │                              │
        ├──────────────────────────────┤
        │         Slab/Exclude         │
        └──────────────────────────────┘
    Args:
        kcl: The KCLayout which will be owned
        additional_info: Add additional key/values to the
            [`KCell.info`][kfactory.kcell.KCell.info]. Can be a static dict
            mapping info name to info value. Or can a callable which takes the straight
            functions' parameters as kwargs and returns a dict with the mapping.
        basename: Overwrite the prefix of the resulting KCell's name. By default
            the KCell will be named 'straight_dbu[...]'.
        cell_kwargs: Additional arguments passed as `@kcl.cell(**cell_kwargs)`.
    """
    if callable(additional_info) and additional_info is not None:
        _additional_info_func: Callable[
            ...,
            dict[str, MetaData],
        ] = additional_info
        _additional_info: dict[str, MetaData] = {}
    else:

        def additional_info_func(
            **kwargs: Any,
        ) -> dict[str, MetaData]:
            return {}

        _additional_info_func = additional_info_func
        _additional_info = additional_info or {}

    @kcl.cell(basename=basename, **cell_kwargs)
    def straight(
        width: kf_types.dbu,
        length: kf_types.dbu,
        layer: kdb.LayerInfo,
        enclosure: LayerEnclosure | None = None,
    ) -> KCell:
        """Waveguide defined in dbu.

            ┌──────────────────────────────┐
            │         Slab/Exclude         │
            ├──────────────────────────────┤
            │                              │
            │             Core             │
            │                              │
            ├──────────────────────────────┤
            │         Slab/Exclude         │
            └──────────────────────────────┘
        Args:
            width: Waveguide width. [dbu]
            length: Waveguide length. [dbu]
            layer: Main layer of the waveguide.
            enclosure: Definition of slab/excludes. [dbu]
        """
        c = kcl.kcell()

        if length < 0:
            logger.critical(
                f"Negative lengths are not allowed {length} as ports"
                " will be inverted. Please use a positive number. Forcing positive"
                " lengths."
            )
            length = -length
        if width < 0:
            logger.critical(
                f"Negative widths are not allowed {width} as ports"
                " will be inverted. Please use a positive number. Forcing positive"
                " lengths."
            )
            width = -width

        if width // 2 * 2 != width:
            raise ValueError("The width (w) must be a multiple of 2 database units")

        li = c.kcl.find_layer(layer)
        c.shapes(li).insert(kdb.Box(0, -width // 2, length, width // 2))
        c.create_port(trans=kdb.Trans(2, False, 0, 0), layer=li, width=width)
        c.create_port(trans=kdb.Trans(0, False, length, 0), layer=li, width=width)

        if enclosure is not None:
            enclosure.apply_minkowski_y(c, layer)
        _info: dict[str, MetaData] = {
            "width_um": width * c.kcl.dbu,
            "length_um": length * c.kcl.dbu,
            "width_dbu": width,
            "length_dbu": length,
        }
        _info.update(
            _additional_info_func(
                width=width, length=length, layer=layer, enclosure=enclosure
            )
        )
        _info.update(_additional_info)
        c.info = Info(**_info)

        c.boundary = c.dbbox()
        c.auto_rename_ports()
        return c

    return straight



"""Factories for generating functions which can produce KCells or VKCells."""

from typing import Protocol

from .. import KCell
from . import bezier, circular, euler, straight, taper, virtual


class StraightFactory(Protocol):
    """Factory Protocol for routing.

    A straight factory must return a KCell with only a width and length given.
    """

    def __call__(self, width: int, length: int) -> KCell:
        """Produces the KCell.

        E.g. in a function this would amount to
        `straight_factory(length=10_000, width=1000)`
        """
        ...


__all__ = [
    "bezier",
    "circular",
    "euler",
    "straight",
    "taper",
    "virtual",
    "StraightFactory",
]



"""Virtual circular cells."""

from ...factories.virtual.circular import virtual_bend_circular_factory
from ...kcell import kcl

__all__ = ["virtual_bend_circular"]

virtual_bend_circular = virtual_bend_circular_factory(kcl)



import kfactory as kf
from functools import partial
from conftest import Layers


def test_virtual_cell() -> None:
    c = kf.VKCell()
    c.shapes(kf.kcl.find_layer(1, 0)).insert(
        kf.kdb.DPolygon([kf.kdb.DPoint(0, 0), kf.kdb.DPoint(1, 0), kf.kdb.DPoint(0, 1)])
    )


def test_virtual_inst(straight: kf.KCell) -> None:
    c = kf.VKCell()
    c << straight


def test_virtual_cell_insert(
    LAYER: Layers, straight: kf.KCell, wg_enc: kf.LayerEnclosure
) -> None:
    c = kf.KCell()

    vc = kf.VKCell("test_virtual_insert")

    e_bend = kf.cells.virtual.euler.virtual_bend_euler(
        width=0.5,
        radius=10,
        layer=LAYER.WG,
        angle=25,
        enclosure=wg_enc,
    )
    e1 = vc << e_bend
    e2 = vc << e_bend
    e3 = vc << e_bend
    e4 = vc << e_bend
    _s = kf.cells.virtual.straight.virtual_straight(
        width=0.5, length=10, layer=LAYER.WG, enclosure=wg_enc
    )
    s = vc << _s

    s.connect("o1", e1, "o2")

    e2.connect("o1", s, "o2")
    e3.connect("o1", e2, "o2")
    e4.connect("o2", e3, "o2")
    s2 = vc << straight
    s2.connect("o1", e4, "o1")

    vi = kf.VInstance(vc)
    vi.insert_into(c)


def test_all_angle_route(LAYER: Layers, wg_enc: kf.LayerEnclosure) -> None:
    bb = [kf.kdb.DPoint(x, y) for x, y in [(0, 0), (500, 0), (250, 200), (500, 250)]]
    c = kf.KCell("test_virtual")
    vc = kf.VKCell("test_all_angle")
    kf.routing.aa.optical.route(
        vc,
        width=5,
        backbone=bb,
        straight_factory=partial(
            kf.cells.virtual.straight.virtual_straight,
            layer=LAYER.WG,
            enclosure=wg_enc,
        ),
        bend_factory=partial(
            kf.cells.virtual.euler.virtual_bend_euler,
            width=5,
            radius=20,
            layer=LAYER.WG,
            enclosure=wg_enc,
        ),
    )
    kf.VInstance(vc, kf.kdb.DCplxTrans()).insert_into(c)



import kfactory as kf
from kfactory.enclosure import extrude_path, extrude_path_dynamic
import numpy as np
from conftest import Layers


@kf.cell
def taper_dyn(
    length: float, width: float, layer: kf.kdb.LayerInfo, enclosure: kf.LayerEnclosure
) -> kf.KCell:
    c = kf.KCell()

    path = [kf.kdb.DPoint(x, 0) for x in range(21)]
    _width = lambda x: width + width * np.sin(x * np.pi / 2)

    extrude_path_dynamic(c, layer, path, _width, enclosure)

    return c


@kf.cell
def taper_static(
    length: float, width: float, layer: kf.kdb.LayerInfo, enclosure: kf.LayerEnclosure
) -> kf.KCell:
    c = kf.KCell()

    path = [kf.kdb.DPoint(x, 0) for x in range(21)]

    _width = [width + np.sin(x * np.pi / 2) for x in [_x / 20 for _x in range(21)]]
    extrude_path_dynamic(c, layer, path, _width, enclosure)

    return c


def test_dynamic_sine_taper(LAYER: Layers, wg_enc: kf.LayerEnclosure) -> None:
    taper_dyn(10, 1, LAYER.WG, wg_enc)


def test_static_sine_taper(LAYER: Layers, wg_enc: kf.LayerEnclosure) -> None:
    taper_static(10, 1, LAYER.WG, wg_enc)


def test_enc_extrude_dyn(LAYER: Layers, wg_enc: kf.LayerEnclosure) -> None:
    width = 10
    layer = LAYER.WG
    enclosure = wg_enc
    c = kf.KCell()

    path = [kf.kdb.DPoint(x, 0) for x in range(21)]
    _width = lambda x: width + width * np.sin(x * np.pi / 2)

    enclosure.extrude_path_dynamic(c, path, layer, _width)


def test_enc_extrude_static(LAYER: Layers, wg_enc: kf.LayerEnclosure) -> None:
    width = 10
    layer = LAYER.WG
    enclosure = wg_enc
    c = kf.KCell()

    path = [kf.kdb.DPoint(x, 0) for x in range(21)]
    _width = [width + np.sin(x * np.pi / 2) for x in [_x / 20 for _x in range(21)]]

    enclosure.extrude_path_dynamic(c, path, layer, _width)



import kfactory as kf
from conftest import Layers


def test_min_width_minkowski(straight: kf.KCell, LAYER: Layers) -> None:
    kf.utils.violations.fix_width_minkowski_tiled(
        c=straight, min_width=500, ref=LAYER.WG
    )


def test_min_spacing_minkowski(straight: kf.KCell, LAYER: Layers) -> None:
    kf.utils.violations.fix_spacing_minkowski_tiled(
        c=straight, min_space=500, ref=LAYER.WG
    )


def test_min_width_and_spacing_minkowski(straight: kf.KCell, LAYER: Layers) -> None:
    kf.utils.violations.fix_width_and_spacing_minkowski_tiled(
        c=straight, min_space=500, min_width=250, ref=LAYER.WG
    )


def test_min_spacing_drc(straight: kf.KCell, LAYER: Layers) -> None:
    kf.utils.violations.fix_spacing_tiled(c=straight, layer=LAYER.WG, min_space=500)


def test_min_spacing_sizing(straight: kf.KCell, LAYER: Layers) -> None:
    kf.utils.violations.fix_spacing_sizing_tiled(
        c=straight, min_space=500, layer=LAYER.WG
    )



# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: -all
#     custom_cell_magics: kql
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.16.2
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Fixing DRC Errors
#
# ## Min space violations
#
# You can fix Min space violations.

# %%
from datetime import datetime
import kfactory as kf

# Define Layers

class LayerInfos(kf.LayerInfos):
    WG: kf.kdb.LayerInfo = kf.kdb.LayerInfo(1,0)
    WGEX: kf.kdb.LayerInfo = kf.kdb.LayerInfo(2,0) # WG Exclude
    CLAD: kf.kdb.LayerInfo = kf.kdb.LayerInfo(3,0) # cladding
    FLOORPLAN: kf.kdb.LayerInfo = kf.kdb.LayerInfo(10,0)

# Make the layout object aware of the new layers:
LAYER = LayerInfos()
kf.kcl.infos = LAYER

# %%
triangle = kf.KCell()
triangle_poly = kf.kdb.DPolygon(
    [kf.kdb.DPoint(0, 10), kf.kdb.DPoint(30, 10), kf.kdb.DPoint(30, 0)]
)
triangle.shapes(triangle.layer(1, 0)).insert(triangle_poly)
triangle

# %%
box = kf.KCell("Box")
box_rect = kf.kdb.DBox(0, 0, 20, 5)
box.shapes(box.kcl.find_layer(1, 0)).insert(box_rect)
box

# %%
c = kf.KCell("fix_accute_angle")
c << triangle
c << box
c

# %%
c = kf.KCell("tiled_test")


d1 = datetime.now()

for i in range(50):
    ellipse = kf.kdb.Polygon.ellipse(kf.kdb.Box(10000, 20000), i * 2)

    x0 = 0
    for j in range(5000, 30000, 500):
        c.shapes(c.kcl.find_layer(1, 0)).insert(
            ellipse.transformed(kf.kdb.Trans(x0, i * 30000))
        )
        c.shapes(c.kcl.find_layer(1, 0)).insert(
            ellipse.transformed(kf.kdb.Trans(x0 + j, i * 30000))
        )

        x0 += 15000

d2 = datetime.now()

c.shapes(c.kcl.layer(2, 0)).insert(
    kf.utils.fix_spacing_tiled(
        c,
        1000,
        c.kcl.infos.WG,
        metrics=kf.kdb.Metrics.Euclidian,
        n_threads=32,
        tile_size=(250, 250),
    )
)

d3 = datetime.now()

print(f"time to draw: {d2-d1}")
print(f"time to clean: {d3-d2}")
print(f"total time: {d3-d1}")

c

# %%
c = kf.KCell()
d1 = datetime.now()

for i in range(50):
    ellipse = kf.kdb.Polygon.ellipse(kf.kdb.Box(10000, 20000), i * 2)

    x0 = 0
    for j in range(5000, 30000, 500):
        c.shapes(c.kcl.layer(1, 0)).insert(
            ellipse.transformed(kf.kdb.Trans(x0, i * 30000))
        )
        c.shapes(c.kcl.layer(1, 0)).insert(
            ellipse.transformed(kf.kdb.Trans(x0 + j, i * 30000))
        )

        x0 += 15000

d2 = datetime.now()

c.shapes(c.kcl.layer(2, 0)).insert(
    kf.utils.fix_spacing_minkowski_tiled(
        c,
        1000,
        c.kcl.infos.WG,
        n_threads=32,
        tile_size=(250, 250),
        smooth=5,
    )
)

d3 = datetime.now()

print(f"time to draw: {d2-d1}")
print(f"time to clean: {d3-d2}")
print(f"total time: {d3-d1}")

c.show()
c.plot()

# %% [markdown]
# ## Dummy fill
#
# To keep density constant you can add dummy fill.

# %%
c = kf.KCell()
c.shapes(kf.kcl.find_layer(1, 0)).insert(
    kf.kdb.DPolygon.ellipse(kf.kdb.DBox(5000, 3000), 512)
)
c.shapes(kf.kcl.find_layer(10, 0)).insert(
    kf.kdb.DPolygon(
        [kf.kdb.DPoint(0, 0), kf.kdb.DPoint(5000, 0), kf.kdb.DPoint(5000, 3000)]
    )
)
c

# %%
fc = kf.KCell()
fc.shapes(fc.kcl.find_layer(2, 0)).insert(kf.kdb.DBox(20, 40))
fc.shapes(fc.kcl.find_layer(3, 0)).insert(kf.kdb.DBox(30, 15))
fc

# %%
import kfactory.utils.fill as fill

# %%
# fill.fill_tiled(c, fc, [(kf.kcl.find_layer(1,0), 0)], exclude_layers = [(kf.kcl.find_layer(10,0), 100), (kf.kcl.find_layer(2,0), 0), (kf.kcl.find_layer(3,0),0)], x_space=5, y_space=5)
fill.fill_tiled(
    c,
    fc,
    [(kf.kcl.infos.WG, 0)],
    exclude_layers=[
        (LAYER.FLOORPLAN, 100),
        (LAYER.WGEX, 0),
        (LAYER.CLAD, 0),
    ],
    x_space=5,
    y_space=5,
)

# %%
c.show()
c.plot()

# %%



"""Virtual cell factories for kfactory.

These are used for cases where any-angle placement is necessary,
for example in all-angle routing.
"""

from . import circular, euler, straight, utils

__all__ = ["circular", "euler", "straight", "utils"]



"""Utilities for automatically routing electrical connections."""

from collections.abc import Callable, Sequence
from typing import Any, Literal

from .. import kdb
from ..conf import logger
from ..kcell import KCell, Port
from ..kf_types import dbu
from .generic import ManhattanRoute
from .generic import route_bundle as route_bundle_generic
from .manhattan import ManhattanRoutePathFunction, route_manhattan, route_smart
from .steps import Step, Straight

__all__ = [
    "route_elec",
    "route_L",
    "route_bundle",
    "route_bundle_dual_rails",
    "route_dual_rails",
    "place_single_wire",
    "place_dual_rails",
]


def route_elec(
    c: KCell,
    p1: Port,
    p2: Port,
    start_straight: int | None = None,
    end_straight: int | None = None,
    route_path_function: ManhattanRoutePathFunction = route_manhattan,
    width: dbu | None = None,
    layer: int | None = None,
    minimum_straight: int | None = None,
) -> None:
    """Connect two ports with a wire.

    A wire is a path object on a usually metal layer.


    Args:
        c: KCell to place the wire in.
        p1: Beginning
        p2: End
        start_straight: Minimum length of straight at start port.
        end_straight: Minimum length of straight at end port.
        route_path_function: Function to calculate the path. Signature:
            `route_path_function(p1, p2, bend90_radius, start_straight,
            end_straight)`
        width: Overwrite the width of the wire. Calculated by the width of the start
            port if `None`.
        layer: Layer to place the wire on. Calculated from the start port if `None`.
        minimum_straight: require a minimum straight
    """
    if width is None:
        width = p1.width
    if layer is None:
        layer = p1.layer
    if start_straight is None:
        start_straight = round(width / 2)
    if end_straight is None:
        end_straight = round(width / 2)

    if minimum_straight is not None:
        start_straight = min(minimum_straight // 2, start_straight)
        end_straight = min(minimum_straight // 2, end_straight)

        pts = route_path_function(
            p1.copy(),
            p2.copy(),
            bend90_radius=minimum_straight,
            start_steps=[Straight(dist=start_straight)],
            end_steps=[Straight(dist=end_straight)],
        )
    else:
        pts = route_path_function(
            p1.copy(),
            p2.copy(),
            bend90_radius=0,
            start_steps=[Straight(dist=start_straight)],
            end_steps=[Straight(dist=end_straight)],
        )

    path = kdb.Path(pts, width)
    c.shapes(layer).insert(path.polygon())


def route_L(
    c: KCell,
    input_ports: list[Port],
    output_orientation: int = 1,
    wire_spacing: dbu = 10000,
) -> list[Port]:
    """Route ports towards a bundle in an L shape.

    This function takes a list of input ports and assume they are oriented in the west.
    The output will be a list of ports that have the same y coordinates.
    The function will produce a L-shape routing to connect input ports to output ports
    without any crossings.
    """
    input_ports.sort(key=lambda p: p.y)

    y_max = input_ports[-1].y
    y_min = input_ports[0].y
    x_max = max(p.x for p in input_ports)

    output_ports = []
    if output_orientation == 1:
        for i, p in enumerate(input_ports[::-1]):
            temp_port = p.copy()
            temp_port.trans = kdb.Trans(
                3, False, x_max - wire_spacing * (i + 1), y_max + wire_spacing
            )

            route_elec(c, p, temp_port)
            temp_port.trans.angle = 1
            output_ports.append(temp_port)
    elif output_orientation == 3:
        for i, p in enumerate(input_ports):
            temp_port = p.copy()
            temp_port.trans = kdb.Trans(
                1, False, x_max - wire_spacing * (i + 1), y_min - wire_spacing
            )
            route_elec(c, p, temp_port)
            temp_port.trans.angle = 3
            output_ports.append(temp_port)
    else:
        raise ValueError(
            "Invalid L-shape routing. Please change output_orientaion to 1 or 3."
        )
    return output_ports


def route_bundle(
    c: KCell,
    start_ports: list[Port],
    end_ports: list[Port],
    separation: dbu,
    start_straights: dbu | list[dbu] = 0,
    end_straights: dbu | list[dbu] = 0,
    place_layer: kdb.LayerInfo | None = None,
    route_width: dbu | None = None,
    bboxes: list[kdb.Box] = [],
    bbox_routing: Literal["minimal", "full"] = "minimal",
    sort_ports: bool = False,
    collision_check_layers: Sequence[kdb.LayerInfo] | None = None,
    on_collision: Literal["error", "show_error"] | None = "show_error",
    waypoints: kdb.Trans | list[kdb.Point] | None = None,
    starts: dbu | list[dbu] | list[Step] | list[list[Step]] = [],
    ends: dbu | list[dbu] | list[Step] | list[list[Step]] = [],
    start_angles: int | list[int] | None = None,
    end_angles: int | list[int] | None = None,
) -> list[ManhattanRoute]:
    r"""Connect multiple input ports to output ports.

    This function takes a list of input ports and assume they are all oriented in the
    same direction (could be any of W, S, E, N). The target ports have the opposite
    orientation, i.e. if input ports are oriented to north, and target ports should
    be oriented to south. The function will produce a routing to connect input ports
    to output ports without any crossings.

    Waypoints will create a front which will create ports in a 1D array. If waypoints
    are a transformation it will be like a point with a direction. If multiple points
    are passed, the direction will be invfered.
    For orientation of 0 degrees it will create the following front for 4 ports:

    ```
          │
          │
          │
          p1 ->
          │
          │
          │


          │
          │
          │
          p2 ->
          │
          │
          │
      ___\waypoint
         /
          │
          │
          │
          p3 ->
          │
          │
          │


          │
          │
          │
          p4 ->
          │
          │
          │
    ```

    Args:
        c: KCell to place the routes in.
        start_ports: List of start ports.
        end_ports: List of end ports.
        separation: Minimum space between wires. [dbu]
        starts: Minimal straight segment after `start_ports`.
        ends: Minimal straight segment before `end_ports`.
        start_straights: Deprecated, use starts instead.
        end_straights: Deprecated, use ends instead.
        place_layer: Override automatic detection of layers with specific layer.
        route_width: Width of the route. If None, the width of the ports is used.
        bboxes: List of boxes to consider. Currently only boxes overlapping ports will
            be considered.
        bbox_routing: "minimal": only route to the bbox so that it can be safely routed
            around, but start or end bends might encroach on the bounding boxes when
            leaving them.
        sort_ports: Automatically sort ports.
        collision_check_layers: Layers to check for actual errors if manhattan routes
            detect potential collisions.
        on_collision: Define what to do on routing collision. Default behaviour is to
            open send the layout of c to klive and open an error lyrdb with the
            collisions. "error" will simply raise an error. None will ignore any error.
        waypoints: Bundle the ports and route them with minimal separation through
            the waypoints. The waypoints can either be a list of at least two points
            or a single transformation. If it's a transformation, the points will be
            routed through it as if it were a tunnel with length 0.
        start_angles: Overwrite the port orientation of all start_ports together
            (single value) or each one (list of values which is as long as start_ports).
        end_angles: Overwrite the port orientation of all start_ports together
            (single value) or each one (list of values which is as long as end_ports).
            If no waypoints are set, the target angles of all ends muts be the same
            (after the steps).
    """
    return route_bundle_generic(
        c=c,
        start_ports=start_ports,
        end_ports=end_ports,
        starts=starts,
        ends=ends,
        routing_function=route_smart,
        routing_kwargs={
            "separation": separation,
            "sort_ports": sort_ports,
            "bbox_routing": bbox_routing,
            "bboxes": list(bboxes),
            "bend90_radius": 0,
            "waypoints": waypoints,
        },
        placer_function=place_single_wire,
        sort_ports=sort_ports,
        on_collision=on_collision,
        collision_check_layers=collision_check_layers,
        start_angles=start_angles,
        end_angles=end_angles,
    )


def route_bundle_dual_rails(
    c: KCell,
    start_ports: list[Port],
    end_ports: list[Port],
    separation: dbu,
    start_straights: dbu | list[dbu] = 0,
    end_straights: dbu | list[dbu] = 0,
    place_layer: kdb.LayerInfo | None = None,
    width_rails: dbu | None = None,
    separation_rails: dbu | None = None,
    bboxes: list[kdb.Box] = [],
    bbox_routing: Literal["minimal", "full"] = "minimal",
    sort_ports: bool = False,
    collision_check_layers: Sequence[kdb.LayerInfo] | None = None,
    on_collision: Literal["error", "show_error"] | None = "show_error",
    waypoints: kdb.Trans | list[kdb.Point] | None = None,
    starts: dbu | list[dbu] | list[Step] | list[list[Step]] = [],
    ends: dbu | list[dbu] | list[Step] | list[list[Step]] = [],
    start_angles: int | list[int] | None = None,
    end_angles: int | list[int] | None = None,
) -> list[ManhattanRoute]:
    r"""Connect multiple input ports to output ports.

    This function takes a list of input ports and assume they are all oriented in the
    same direction (could be any of W, S, E, N). The target ports have the opposite
    orientation, i.e. if input ports are oriented to north, and target ports should
    be oriented to south. The function will produce a routing to connect input ports
    to output ports without any crossings.

    Waypoints will create a front which will create ports in a 1D array. If waypoints
    are a transformation it will be like a point with a direction. If multiple points
    are passed, the direction will be invfered.
    For orientation of 0 degrees it will create the following front for 4 ports:

    ```
          │
          │
          │
          p1 ->
          │
          │
          │


          │
          │
          │
          p2 ->
          │
          │
          │
      ___\waypoint
         /
          │
          │
          │
          p3 ->
          │
          │
          │


          │
          │
          │
          p4 ->
          │
          │
          │
    ```

    Args:
        c: KCell to place the routes in.
        start_ports: List of start ports.
        end_ports: List of end ports.
        separation: Minimum space between wires. [dbu]
        starts: Minimal straight segment after `start_ports`.
        ends: Minimal straight segment before `end_ports`.
        start_straights: Deprecated, use starts instead.
        end_straights: Deprecated, use ends instead.
        place_layer: Override automatic detection of layers with specific layer.
        width_rails: Total width of the rails.
        separation_rails: Separation between the two rails.
        bboxes: List of boxes to consider. Currently only boxes overlapping ports will
            be considered.
        bbox_routing: "minimal": only route to the bbox so that it can be safely routed
            around, but start or end bends might encroach on the bounding boxes when
            leaving them.
        sort_ports: Automatically sort ports.
        collision_check_layers: Layers to check for actual errors if manhattan routes
            detect potential collisions.
        on_collision: Define what to do on routing collision. Default behaviour is to
            open send the layout of c to klive and open an error lyrdb with the
            collisions. "error" will simply raise an error. None will ignore any error.
        waypoints: Bundle the ports and route them with minimal separation through
            the waypoints. The waypoints can either be a list of at least two points
            or a single transformation. If it's a transformation, the points will be
            routed through it as if it were a tunnel with length 0.
        start_angles: Overwrite the port orientation of all start_ports together
            (single value) or each one (list of values which is as long as start_ports).
        end_angles: Overwrite the port orientation of all start_ports together
            (single value) or each one (list of values which is as long as end_ports).
            If no waypoints are set, the target angles of all ends muts be the same
            (after the steps).
    """
    if start_straights is not None:
        logger.warning("start_straights is deprecated. Use `starts` instead.")
        starts = start_straights
    if end_straights is not None:
        logger.warning("end_straights is deprecated. Use `starts` instead.")
        ends = end_straights
    return route_bundle_generic(
        c=c,
        start_ports=start_ports,
        end_ports=end_ports,
        routing_function=route_smart,
        starts=starts,
        ends=ends,
        routing_kwargs={
            "separation": separation,
            "sort_ports": sort_ports,
            "bbox_routing": bbox_routing,
            "bboxes": list(bboxes),
            "bend90_radius": 0,
            "waypoints": waypoints,
        },
        placer_function=place_dual_rails,
        placer_kwargs={
            "separation_rails": separation_rails,
        },
        sort_ports=sort_ports,
        on_collision=on_collision,
        collision_check_layers=collision_check_layers,
        start_angles=start_angles,
        end_angles=end_angles,
    )


def route_dual_rails(
    c: KCell,
    p1: Port,
    p2: Port,
    start_straight: dbu | None = None,
    end_straight: dbu | None = None,
    route_path_function: Callable[..., list[kdb.Point]] = route_manhattan,
    width: dbu | None = None,
    hole_width: dbu | None = None,
    layer: int | None = None,
) -> None:
    """Connect ports with a dual-wire rail.

    Args:
        c: KCell to place the connection in.
        p1: Start port.
        p2: End port.
        start_straight: Minimum straight after the start port.
        end_straight: Minimum straight before end port.
        route_path_function: Function to calculate the path. Signature:
            `route_path_function(p1, p2, bend90_radius, start_straight,
            end_straight)`
        width: Width of the rail (total). [dbu]
        hole_width: Width of the space between the rails. [dbu]
        layer: layer to place the rail in.
    """
    _width = width or p1.width
    _hole_width = hole_width or p1.width // 2
    _layer = layer or p1.layer

    pts = route_path_function(
        p1.copy(),
        p2.copy(),
        bend90_radius=0,
        start_straight=start_straight,
        end_straight=end_straight,
    )

    path = kdb.Path(pts, _width)
    hole_path = kdb.Path(pts, _hole_width)
    final_poly = kdb.Region(path.polygon()) - kdb.Region(hole_path.polygon())
    c.shapes(_layer).insert(final_poly)


def place_single_wire(
    c: KCell,
    p1: Port,
    p2: Port,
    pts: Sequence[kdb.Point],
    route_width: dbu | None = None,
    layer_info: kdb.LayerInfo | None = None,
    **kwargs: Any,
) -> ManhattanRoute:
    """Placer function for a single wire.

    Args:
        c: KCell to place the route in.
        p1: Start port.
        p2: End port.
        pts: Route backbone.
        route_width: Overwrite automatic detection of wire width.
        layer_info: Place on a specific layer. Otherwise, use
            `p1.layer_info`.
        width: Place a route with a specific width. Otherwise, use
            `p2.width`.
        kwargs: Compatibility for type checkers. Throws an error if not empty.
    """
    if layer_info is None:
        layer_info = p1.layer_info
    if route_width is None:
        route_width = p1.width
    if kwargs:
        raise ValueError(
            f"Additional kwargs aren't supported in route_single_wire {kwargs=}"
        )

    shape = (
        c.shapes(c.layer(layer_info))
        .insert(kdb.Path(pts, width=route_width).polygon())
        .polygon
    )

    _length = 0.0
    pt1 = pts[0]
    for pt2 in pts[1:]:
        _length += (pt2 - pt1).length()

    return ManhattanRoute(
        backbone=pts,
        start_port=p1,
        end_port=p2,
        taper_length=0,
        bend90_radius=0,
        polygons={layer_info: [shape]},
        instances=[],
        length=round(_length),
        length_straight=round(_length),
    )


def place_dual_rails(
    c: KCell,
    p1: Port,
    p2: Port,
    pts: Sequence[kdb.Point],
    route_width: dbu | None = None,
    layer_info: kdb.LayerInfo | None = None,
    separation_rails: dbu | None = None,
    **kwargs: Any,
) -> ManhattanRoute:
    """Placer function for a single wire.

    Args:
        c: KCell to place the route in.
        p1: Start port.
        p2: End port.
        pts: Route backbone.
        route_width: Overwrite automatic detection of wire width.
            Total width of all rails.
        layer_info: Place on a specific layer. Otherwise, use
            `p1.layer_info`.
        width_rails: Total width of the rails.
        separation_rails: Separation between the two rails.
        kwargs: Compatibility for type checkers. Throws an error if not empty.
    """
    if kwargs:
        raise ValueError(
            f"Additional kwargs aren't supported in route_dual_rails {kwargs=}"
        )
    if layer_info is None:
        layer_info = p1.layer_info
    if route_width is None:
        route_width = p1.width
    if separation_rails is None:
        raise ValueError("Must specify a separation between the two rails.")
    if separation_rails >= route_width:
        raise ValueError(f"{separation_rails=} must be smaller than the {route_width}")

    region = kdb.Region(kdb.Path(pts, route_width)) - kdb.Region(
        kdb.Path(pts, separation_rails)
    )

    shapes = [
        c.shapes(c.layer(layer_info)).insert(region[0]).polygon,
        c.shapes(c.layer(layer_info)).insert(region[1]).polygon,
    ]

    return ManhattanRoute(
        backbone=pts,
        start_port=p1,
        end_port=p2,
        taper_length=0,
        bend90_radius=0,
        polygons={layer_info: shapes},
        instances=[],
    )



"""CLI interface for kfactory.

Use `kf --help` for more info.
"""

from typing import Annotated

# import click
import typer

from .. import __version__
from .build import build, show

app = typer.Typer(name="kf")


app.command()(build)
app.command()(show)


@app.callback(invoke_without_command=True)
def version_callback(
    version: Annotated[
        bool, typer.Option("--version", "-V", help="Show version of the CLI")
    ] = False,
) -> None:
    """Show the version of the cli."""
    if version:
        print(f"KFactory CLI Version: {__version__}")
        raise typer.Exit()



from __future__ import annotations
from typing import TYPE_CHECKING
from pathlib import Path

if TYPE_CHECKING:
    import kfactory


def show(
    layout: kfactory.KCLayout | kfactory.KCell | Path | str,
    *,
    lyrdb: kfactory.rdb.ReportDatabase | Path | str | None = None,
    l2n: kfactory.kdb.LayoutToNetlist | Path | str | None = None,
    keep_position: bool = True,
    save_options: kfactory.kdb.SaveLayoutOptions | None = None,
    use_libraries: bool = True,
    library_save_options: kfactory.kdb.SaveLayoutOptions | None = None,
) -> None:
    import kfactory as kf

    if save_options is None:
        save_options = kf.save_layout_options()
    if library_save_options is None:
        library_save_options = kf.save_layout_options()



"""CLI interface for kfactory.

Use `kf --help` for more info.
"""
# ruff: noqa: UP007

import importlib
import os
import runpy
import sys
from enum import Enum
from pathlib import Path
from typing import Annotated, Optional

import git
import typer

from ..conf import logger
from ..kcell import KCell, kcls, save_layout_options
from ..kcell import show as kfshow

__all__ = ["show", "build"]


def show(
    file: Path,
) -> None:
    """Show a GDS or OAS file in KLayout through klive."""
    path = Path(file)
    logger.debug("Path = {}", path.resolve())
    if not path.exists():
        logger.critical("{file} does not exist, exiting", file=file)
        return
    if not path.is_file():
        logger.critical("{file} is not a file, exiting", file=file)
        return
    if not os.access(path, os.R_OK):
        logger.critical("No permission to read file {file}, exiting", file=file)
        return
    kfshow(path, use_libraries=True)


class LayoutSuffix(str, Enum):
    gds = "gds"
    gdsgz = "gds.gz"
    oas = "oas"


def build(
    build_ref: Annotated[
        str,
        typer.Argument(
            default=...,
            help="The file|module|function to execute:\n"
            "Accepted Formats:\n- /path/to/file.py::cell_function_name\n"
            "- module.submodule::cell_function_name\n"
            "- /path/to/file.py\n"
            "- module.submodule",
        ),
    ],
    func_kwargs: Annotated[
        Optional[list[str]],  # noqa: UP007
        typer.Argument(
            help="Arguments used for --type function."
            " Doesn't have any influence for other types"
        ),
    ] = None,
    show: Annotated[
        bool, typer.Option(help="Show the file through klive in KLayout")
    ] = True,
    library: Annotated[
        Optional[list[str]],
        typer.Option(
            help="Library(s) used by the main layout file. Only works for functions,"
            " not for '__main__' modules"
        ),
    ] = None,
    suffix: Annotated[
        LayoutSuffix, typer.Option(help="Format of the layout files")
    ] = LayoutSuffix.oas,
    write_full: Annotated[
        bool,
        typer.Option(
            help="Write the gds with full library support. Uses libraries passed with"
            " --library. Only works in function mode not on modules"
        ),
    ] = True,
    write_static: Annotated[
        bool,
        typer.Option(
            help="Write a layout where all cells are converted to static (non-library)"
            " cells."
        ),
    ] = False,
    write_nocontext: Annotated[
        bool,
        typer.Option(
            help="Write the layout without any meta infos. This is useful for "
            "submitting the GDS to fabs."
        ),
    ] = False,
    max_vertex_count: Annotated[
        int, typer.Option(help="Maximum number of vertices per polygon.")
    ] = 4000,
    max_cellname_length: Annotated[
        int, typer.Option(help="Maximum number of characters in a cell name.")
    ] = 99,
) -> None:
    """Run a python modules __main__ or a function if specified."""
    path = sys.path.copy()
    sys.path.append(os.getcwd())
    saveopts = save_layout_options()
    saveopts.gds2_max_cellname_length = max_cellname_length
    saveopts.gds2_max_vertex_count = max_vertex_count

    build_split = build_ref.rsplit("::", 1)
    if len(build_split) == 1:
        mod_file = build_split[0]
        func: str | None = None
    else:
        mod_file, func = build_split

    if mod_file.endswith(".py"):
        file = Path(mod_file).expanduser().resolve()
        if not file.is_file():
            raise ImportError(f"File {file} does not exist")
        sys.path.insert(0, str(file.parent))
        if func is not None:
            logger.debug(f"{mod_file=},{func=}")
            try:
                spec = importlib.util.find_spec(str(file.name.removesuffix(".py")))
                if spec is None or spec.loader is None:
                    raise ImportError
                _mod = importlib.util.module_from_spec(spec)
                sys.modules[file.stem] = _mod
                spec.loader.exec_module(_mod)
                sys.path.pop(0)
                kwargs = {}

                old_arg = ""
                if func_kwargs is not None:
                    for i, kwarg in enumerate(func_kwargs):
                        if i % 2:
                            try:
                                value: int | float | str = int(kwarg)
                            except ValueError:
                                try:
                                    value = float(kwarg)
                                except ValueError:
                                    value = kwarg
                            kwargs[old_arg] = value
                        else:
                            old_arg = kwarg

                cell = getattr(_mod, func)(**kwargs)
                if isinstance(cell, KCell):
                    try:
                        repo = git.repo.Repo(".", search_parent_directories=True)
                    except git.InvalidGitRepositoryError:
                        pass
                    else:
                        wtd = repo.working_tree_dir
                        if wtd is not None:
                            root = Path(wtd) / "build/gds"
                            root.mkdir(parents=True, exist_ok=True)
                        else:
                            root = Path()
                    if show:
                        cell.show()
                    if write_full:
                        if library is not None:
                            for lib in library:
                                kcls[lib].write(root / f"{lib}.{suffix.value}")
                        cell.write(
                            root / f"{cell.name}.{suffix.value}",
                            save_options=saveopts,
                        )
                    if write_static:
                        cell.write(
                            root / f"{cell.name}_STATIC.{suffix.value}",
                            convert_external_cells=True,
                            save_options=saveopts,
                        )
                    if write_nocontext:
                        saveopts.write_context_info = False
                        cell.write(
                            root / f"{cell.name}_NOCONT.{suffix.value}",
                            save_options=saveopts,
                            convert_external_cells=True,
                        )

            except ImportError:
                logger.critical(
                    f"Couldn't import function '{func}' from module '"
                    f"{file.with_suffix('')}"
                    "'"
                )
        else:
            runpy.run_path(mod_file, run_name="__main__")
            sys.path.pop(0)
    else:
        sys.path.insert(0, ".")
        if func:
            logger.debug(f"{mod_file=},{func=}")
            try:
                _mod = importlib.import_module(mod_file)
                kwargs = {}

                old_arg = ""
                if func_kwargs is not None:
                    for i, kwarg in enumerate(func_kwargs):
                        if i % 2:
                            try:
                                value = int(kwarg)
                            except ValueError:
                                try:
                                    value = float(kwarg)
                                except ValueError:
                                    value = kwarg
                            kwargs[old_arg] = value
                        else:
                            old_arg = kwarg

                cell = getattr(_mod, func)(**kwargs)
                if isinstance(cell, KCell):
                    try:
                        repo = git.repo.Repo(".", search_parent_directories=True)
                    except git.InvalidGitRepositoryError:
                        pass
                    else:
                        wtd = repo.working_tree_dir
                        if wtd is not None:
                            root = Path(wtd) / "build/gds"
                            root.mkdir(parents=True, exist_ok=True)
                        else:
                            root = Path()
                    if show:
                        cell.show()
                    if write_full:
                        if library is not None:
                            for lib in library:
                                kcls[lib].write(root / f"{lib}.{suffix.value}")
                        cell.write(
                            root / f"{cell.name}.{suffix.value}",
                            save_options=saveopts,
                        )
                    if write_static:
                        cell.write(
                            root / f"{cell.name}_STATIC.{suffix.value}",
                            convert_external_cells=True,
                            save_options=saveopts,
                        )
                    if write_nocontext:
                        saveopts = save_layout_options()
                        saveopts.write_context_info = False
                        cell.write(
                            root / f"{cell.name}_NOCONT.{suffix.value}",
                            save_options=saveopts,
                            convert_external_cells=True,
                        )
            except ImportError:
                logger.critical(
                    f"Couldn't import function '{func}' from module '{mod_file}'"
                )
        else:
            runpy.run_module(mod_file, run_name="__main__")
        sys.path.pop(0)

    sys.path = path
